import numpy as np

from xarray import Dataset, DataArray
from xarray.testing import assert_equal
import pytest

from xbout.geometries import (
    apply_geometry,
    register_geometry,
    UnregisteredGeometryError,
    REGISTERED_GEOMETRIES,
)


class TestGeometryRegistration:
    def test_unregistered_geometry(self):
        with pytest.raises(
            UnregisteredGeometryError, match="tesseract is not a registered geometry"
        ):
            apply_geometry(ds=Dataset(), geometry_name="tesseract")

    def test_register_new_geometry(self):
        @register_geometry(name="Schwarzschild")
        def add_schwarzschild_coords(ds, coordinates=None):
            ds["event_horizon"] = 4.0
            return ds

        assert "Schwarzschild" in REGISTERED_GEOMETRIES.keys()

        original = Dataset()
        original["dy"] = DataArray(np.ones((3, 4)), dims=("x", "y"))
        metadata = {
            "bout_tdim": "t",
            "bout_xdim": "x",
            "bout_ydim": "y",
            "bout_zdim": "z",
            "keep_xboundaries": True,
            "keep_yboundaries": True,
        }
        original.attrs["metadata"] = metadata
        original["dy"].attrs["metadata"] = metadata
        updated = apply_geometry(ds=original, geometry_name="Schwarzschild")
        assert_equal(updated["event_horizon"], DataArray(4.0))

        # clean up
        del REGISTERED_GEOMETRIES["Schwarzschild"]



import numpy as np
import xarray as xr


# TODO write a decorator to add functions as accessors to dataarrays?


def _rms_gufunc(x):
    """Generalized ufunc to calculate root mean squared value of an array."""
    squares = np.square(x)
    return np.sqrt(np.mean(squares, axis=-1))


def rms(da, dim=None, dask="parallelized", keep_attrs=True):
    """
    Reduces a dataarray by calculating the root mean square along the dimension
    dim.
    """

    # TODO If dim is None then take the root mean square along all dimensions?
    if dim is None:
        raise ValueError("Must supply a dimension along which to calculate rms")

    rms = xr.apply_ufunc(
        _rms_gufunc,
        da,
        input_core_dims=[[dim]],
        dask=dask,
        output_dtypes=[da.dtype],
        keep_attrs=keep_attrs,
    )

    # Return the name of the da as variable_rms
    rms.name = str(da.name) + "_rms"

    return rms



from copy import deepcopy
from boutdata.data import BoutOptionsFile
from gelidum import freeze
import inspect
import numpy as np
from pathlib import Path
from xarray import DataArray, Dataset


# Note, MYPE, PE_XIND and PE_YIND not included, since they are different for each
# processor and so are dropped when loading datasets.
METADATA_VARS = [
    "BOUT_VERSION",
    "NXPE",
    "NYPE",
    "NZPE",
    "MXG",
    "MYG",
    "MZG",
    "nx",
    "ny",
    "nz",
    "MZ",
    "MXSUB",
    "MYSUB",
    "MZSUB",
    "hist_hi",
    "iteration",
    "ixseps1",
    "ixseps2",
    "jyseps1_1",
    "jyseps1_2",
    "jyseps2_1",
    "jyseps2_2",
    "ny_inner",
    "tt",
    "zperiod",
    "ZMIN",
    "ZMAX",
    "use_metric_3d",
]


def load_example_input(name, *, nx=None, ny=None, nz=None):
    # Find input file in the 'inputs' subdirectory
    path = Path(__file__).parent.joinpath("inputs", name)

    return BoutOptionsFile(path, nx=nx, ny=ny, nz=nz)


def set_geometry_from_input_file(ds, name):
    ds = ds.copy()  # do not modify existing ds

    options = load_example_input(
        name, nx=ds.metadata["nx"], ny=ds.metadata["ny"], nz=ds.metadata["nz"]
    )

    # Set nx and ny in the options in case these are needed for expressions, as this is
    # not done automatically by BoutOptionsFile
    options["mesh:nx"] = ds.metadata["nx"]
    options["mesh:ny"] = ds.metadata["ny"]

    if not ds.metadata["MXG"] == options._keys.get("MXG", 2):
        raise ValueError(
            f"MXG does not match, {ds.metadata['MXG']} in ds and "
            f"{options._keys.get('MXG', 2)} in options"
        )
    if not ds.metadata["MYG"] == options._keys.get("MYG", 2):
        raise ValueError(
            f"MYG does not match, {ds.metadata['MYG']} in ds and "
            f"{options._keys.get('MYG', 2)} in options"
        )

    if ds.metadata["keep_xboundaries"] or ds.metadata["MXG"] == 0:
        slicex = slice(None)
    else:
        mxg = ds.metadata["MXG"]
        slicex = slice(mxg, -mxg)

    if ds.metadata["keep_yboundaries"] or ds.metadata["MYG"] == 0:
        slicey = slice(None)
    else:
        myg = ds.metadata["MYG"]
        slicey = slice(myg, -myg)

    slices = (slicex, slicey)

    # Replace values of fields already created by bout_xyt_example_files
    shape_2d = ds["g11"].shape
    for v in [
        "g11",
        "g22",
        "g33",
        "g12",
        "g13",
        "g23",
        "g_11",
        "g_22",
        "g_33",
        "g_12",
        "g_13",
        "g_23",
        "J",
        "Bxy",
        "zShift",
        "dx",
        "dy",
    ]:
        for location in ["CELL_CENTRE", "CELL_XLOW", "CELL_YLOW", "CELL_ZLOW"]:
            suffix = "" if location == "CELL_CENTRE" else f"_{location}"
            # Need all arrays returned from options.evaluate() to be the right shape.
            # Recommend adding '0*x' or '0*y' in the input file expressions if the
            # expression would be 1d otherwise.
            ds[v + suffix] = ds[v].copy(
                data=np.broadcast_to(
                    options.evaluate(f"mesh:{v}", location=location).squeeze(axis=2)[
                        slices
                    ],
                    shape_2d,
                )
            )

    # Set dz as it would be calculated by BOUT++ (but don't support zmin, zmax or
    # zperiod here)
    if "dz" in options["mesh"]:
        ds["dz"] = options.evaluate_scalar("mesh:dz")
    else:
        ds["dz"] = 2.0 * np.pi / ds.metadata["nz"]

    # Add extra fields needed by "toroidal" geometry
    for v in ["Rxy", "Zxy", "psixy"]:
        for location in ["CELL_CENTRE", "CELL_XLOW", "CELL_YLOW", "CELL_ZLOW"]:
            suffix = "" if location == "CELL_CENTRE" else f"_{location}"
            # Need all arrays returned from options.evaluate() to be the right shape.
            # Recommend adding '0*x' or '0*y' in the input file expressions if the
            # expression would be 1d otherwise.
            ds[v + suffix] = ds["g11"].copy(
                data=np.broadcast_to(
                    options.evaluate(f"mesh:{v}", location=location).squeeze(axis=2)[
                        slices
                    ],
                    shape_2d,
                )
            )

    # Set fields that don't have to be in input files to NaN
    for v in ["G1", "G2", "G3"]:
        if v in ds:
            ds[v] = ds[v].copy(data=np.broadcast_to(float("nan"), ds[v].shape))

    return ds, options


def _get_kwargs(ignore=None):
    """
    Get the arguments of a function as a frozendict. Extended version of code from here:
    https://stackoverflow.com/a/65927265

    Parameters
    ----------
    ignore : str or Sequence of str
        Arguments to drop when constructing the returned frozendict
    """
    frame = inspect.currentframe().f_back
    keys, _, _, values = inspect.getargvalues(frame)
    kwargs = {}
    keys_to_ignore = ["self"]
    if ignore is not None:
        if isinstance(ignore, str):
            ignore = [ignore]
        keys_to_ignore += ignore
    for key in keys:
        if key not in keys_to_ignore:
            value = values[key]
            kwargs[key] = value

    return freeze(kwargs)


def create_bout_ds_list(
    prefix,
    lengths=(6, 2, 4, 7),
    nxpe=4,
    nype=2,
    nt=1,
    guards={},
    topology="core",
    syn_data_type="random",
    squashed=False,
    bout_v5=False,
    metric_3D=False,
):
    """
    Mocks up a set of BOUT-like datasets.

    Structured as though they were produced by a x-y parallelised run with multiple restarts.
    """

    if nt != 1:
        raise ValueError(
            "nt > 1 means the time dimension is split over several "
            + "directories. This is not implemented yet."
        )

    file_list = []
    ds_list = []
    for j in range(nype):
        for i in range(nxpe):
            num = i + nxpe * j
            filename = prefix + "." + str(num) + ".nc"
            file_list.append(filename)

            ds = create_bout_ds(
                syn_data_type=syn_data_type,
                num=num,
                lengths=lengths,
                nxpe=nxpe,
                nype=nype,
                xproc=i,
                yproc=j,
                guards=guards,
                topology=topology,
                squashed=squashed,
                bout_v5=bout_v5,
                metric_3D=metric_3D,
            )
            ds_list.append(ds)

    return ds_list, file_list


_create_bout_ds_cache = {}


def create_bout_ds(
    syn_data_type="random",
    lengths=(6, 2, 4, 7),
    num=0,
    nxpe=1,
    nype=1,
    xproc=0,
    yproc=0,
    guards=None,
    topology="core",
    squashed=False,
    bout_v5=False,
    metric_3D=False,
):
    call_args = _get_kwargs()

    try:
        # Has been called with the same signature before, just return the cached result
        return deepcopy(_create_bout_ds_cache[call_args])
    except KeyError:
        pass

    if metric_3D and not bout_v5:
        raise ValueError("3D metric requires BOUT++ v5")

    if guards is None:
        guards = {}

    # Set the shape of the data in this dataset
    t_length, x_length, y_length, z_length = lengths
    mxg = guards.get("x", 0)
    myg = guards.get("y", 0)
    x_length += 2 * mxg
    y_length += 2 * myg

    # calculate global nx, ny and nz
    nx = nxpe * lengths[1] + 2 * mxg
    ny = nype * lengths[2]
    nz = 1 * lengths[3]

    if squashed and "double-null" in topology:
        ny = ny + 2 * myg
        y_length = y_length + 2 * myg
    shape = (t_length, x_length, y_length, z_length)

    # Fill with some kind of synthetic data
    if syn_data_type == "random":
        # Each dataset contains unique random noise
        np.random.seed(seed=num)
        data = np.random.randn(*shape)
    elif syn_data_type == "linear":
        # Variables increase linearly across entire domain
        data = DataArray(-np.ones(shape), dims=("t", "x", "y", "z"))

        t_array = DataArray(
            (nx - 2 * mxg) * ny * nz * np.arange(t_length, dtype=float), dims="t"
        )
        x_array = DataArray(
            ny * nz * (xproc * lengths[1] + np.arange(lengths[1], dtype=float)),
            dims="x",
        )
        y_array = DataArray(
            nz * (yproc * lengths[2] + np.arange(lengths[2], dtype=float)), dims="y"
        )
        z_array = DataArray(np.arange(z_length, dtype=float), dims="z")

        data[:, mxg : x_length - mxg, myg : lengths[2] + myg, :] = (
            t_array + x_array + y_array + z_array
        )
    elif syn_data_type == "stepped":
        # Each dataset contains a different number depending on the filename
        data = np.ones(shape) * num
    elif isinstance(syn_data_type, int):
        data = np.ones(shape) * syn_data_type
    else:
        raise ValueError("Not a recognised choice of type of synthetic bout data.")

    T = DataArray(data, dims=["t", "x", "y", "z"])
    n = DataArray(data, dims=["t", "x", "y", "z"])
    S = DataArray(data[:, :, :, 0], dims=["t", "x", "y"])
    for v in [n, T]:
        v.attrs["direction_y"] = "Standard"
        v.attrs["cell_location"] = "CELL_CENTRE"
        v.attrs["direction_z"] = "Standard"
    for v in [S]:
        v.attrs["direction_y"] = "Standard"
        v.attrs["cell_location"] = "CELL_CENTRE"
        v.attrs["direction_z"] = "Average"
    ds = Dataset({"n": n, "T": T, "S": S})

    # BOUT_VERSION needed to deal with backwards incompatible changes:
    #
    # - v3 and earlier: number of points in z is MZ-1
    # - v4 and later: number of points in z is MZ
    # - v5 and later: metric components can be either 2D or 3D
    # - v5 and later: dz changed to be a Field2D/3D
    ds["BOUT_VERSION"] = 5.0 if bout_v5 else 4.3
    ds["use_metric_3d"] = int(metric_3D)

    # Include grid data
    ds["NXPE"] = nxpe
    ds["NYPE"] = nype
    ds["NZPE"] = 1
    ds["PE_XIND"] = xproc
    ds["PE_YIND"] = yproc
    ds["MYPE"] = num

    ds["MXG"] = mxg
    ds["MYG"] = myg
    ds["MZG"] = 0
    ds["nx"] = nx
    ds["ny"] = ny
    ds["nz"] = nz
    ds["MZ"] = 1 * lengths[3]
    if squashed:
        ds["MXSUB"] = lengths[1] // nxpe
        ds["MYSUB"] = lengths[2] // nype
        ds["MZSUB"] = lengths[3]
    else:
        ds["MXSUB"] = lengths[1]
        ds["MYSUB"] = lengths[2]
        ds["MZSUB"] = lengths[3]

    MYSUB = lengths[2]

    if topology == "core":
        ds["ixseps1"] = nx
        ds["ixseps2"] = nx
        ds["jyseps1_1"] = -1
        ds["jyseps2_1"] = ny // 2 - 1
        ds["jyseps1_2"] = ny // 2 - 1
        ds["jyseps2_2"] = ny - 1
        ds["ny_inner"] = ny // 2
    elif topology == "sol":
        ds["ixseps1"] = 0
        ds["ixseps2"] = 0
        ds["jyseps1_1"] = -1
        ds["jyseps2_1"] = ny // 2 - 1
        ds["jyseps1_2"] = ny // 2 - 1
        ds["jyseps2_2"] = ny - 1
        ds["ny_inner"] = ny // 2
    elif topology == "limiter":
        ds["ixseps1"] = nx // 2
        ds["ixseps2"] = nx
        ds["jyseps1_1"] = -1
        ds["jyseps2_1"] = ny // 2 - 1
        ds["jyseps1_2"] = ny // 2 - 1
        ds["jyseps2_2"] = ny - 1
        ds["ny_inner"] = ny // 2
    elif topology == "xpoint":
        if nype < 4 and not squashed:
            raise ValueError(f"Not enough processors for xpoint topology: nype={nype}")
        ds["ixseps1"] = nx // 2
        ds["ixseps2"] = nx // 2
        ds["jyseps1_1"] = MYSUB - 1
        ny_inner = 2 * MYSUB
        ds["ny_inner"] = ny_inner
        ds["jyseps2_1"] = MYSUB - 1
        ds["jyseps1_2"] = ny - MYSUB - 1
        ds["jyseps2_2"] = ny - MYSUB - 1
    elif topology == "single-null":
        if nype < 3 and not squashed:
            raise ValueError(f"Not enough processors for xpoint topology: nype={nype}")
        ds["ixseps1"] = nx // 2
        ds["ixseps2"] = nx
        ds["jyseps1_1"] = MYSUB - 1
        ds["jyseps2_1"] = ny // 2 - 1
        ds["jyseps1_2"] = ny // 2 - 1
        ds["jyseps2_2"] = ny - MYSUB - 1
        ds["ny_inner"] = ny // 2
    elif topology == "connected-double-null":
        if nype < 6 and not squashed:
            raise ValueError(
                "Not enough processors for connected-double-null topology: "
                f"nype={nype}"
            )
        ds["ixseps1"] = nx // 2
        ds["ixseps2"] = nx // 2
        ds["jyseps1_1"] = MYSUB - 1
        ny_inner = 3 * MYSUB
        ds["ny_inner"] = ny_inner
        ds["jyseps2_1"] = ny_inner - MYSUB - 1
        ds["jyseps1_2"] = ny_inner + MYSUB - 1
        ds["jyseps2_2"] = ny - MYSUB - 1
    elif topology == "lower-disconnected-double-null":
        if nype < 6 and not squashed:
            raise ValueError(
                "Not enough processors for lower-disconnected-double-null "
                f"topology: nype={nype}"
            )
        ds["ixseps1"] = nx // 2
        ds["ixseps2"] = nx // 2 + 4
        if ds["ixseps2"] >= nx:
            raise ValueError(
                "Not enough points in the x-direction. ixseps2="
                f'{ds["ixseps2"]} > nx={nx}'
            )
        ds["jyseps1_1"] = MYSUB - 1
        ny_inner = 3 * MYSUB
        ds["ny_inner"] = ny_inner
        ds["jyseps2_1"] = ny_inner - MYSUB - 1
        ds["jyseps1_2"] = ny_inner + MYSUB - 1
        ds["jyseps2_2"] = ny - MYSUB - 1
    elif topology == "upper-disconnected-double-null":
        if nype < 6 and not squashed:
            raise ValueError(
                "Not enough processors for upper-disconnected-double-null "
                f"topology: nype={nype}"
            )
        ds["ixseps2"] = nx // 2
        ds["ixseps1"] = nx // 2 + 4
        if ds["ixseps2"] >= nx:
            raise ValueError(
                "Not enough points in the x-direction. ixseps2="
                f'{ds["ixseps2"]} > nx={nx}'
            )
        ds["jyseps1_1"] = MYSUB - 1
        ny_inner = 3 * MYSUB
        ds["ny_inner"] = ny_inner
        ds["jyseps2_1"] = ny_inner - MYSUB - 1
        ds["jyseps1_2"] = ny_inner + MYSUB - 1
        ds["jyseps2_2"] = ny - MYSUB - 1
    else:
        raise ValueError(f"Unrecognised topology={topology}")

    if metric_3D:
        one = DataArray(np.ones((x_length, y_length, z_length)), dims=["x", "y", "z"])
        zero = DataArray(np.zeros((x_length, y_length, z_length)), dims=["x", "y", "z"])
    else:
        one = DataArray(np.ones((x_length, y_length)), dims=["x", "y"])
        zero = DataArray(np.zeros((x_length, y_length)), dims=["x", "y"])

    ds["zperiod"] = 1
    ds["ZMIN"] = 0.0
    ds["ZMAX"] = 1.0
    ds["g11"] = one
    ds["g22"] = one
    ds["g33"] = one
    ds["g12"] = zero
    ds["g13"] = zero
    ds["g23"] = zero
    ds["g_11"] = one
    ds["g_22"] = one
    ds["g_33"] = one
    ds["g_12"] = zero
    ds["g_13"] = zero
    ds["g_23"] = zero
    ds["G1"] = zero
    ds["G2"] = zero
    ds["G3"] = zero
    ds["J"] = one
    ds["Bxy"] = one
    ds["zShift"] = zero

    ds["dx"] = 0.5 * one
    ds["dy"] = 2.0 * one
    if bout_v5:
        ds["dz"] = 2.0 * one * np.pi / nz
    else:
        ds["dz"] = 2.0 * np.pi / nz

    ds["iteration"] = t_length - 1
    ds["hist_hi"] = t_length - 1
    ds["t_array"] = DataArray(np.arange(t_length, dtype=float) * 10.0, dims="t")
    ds["tt"] = ds["t_array"][-1]

    # xarray adds this encoding when opening a file. Emulate here as it may be used to
    # get the file number
    ds.encoding["source"] = f"BOUT.dmp.{num}.nc"

    _create_bout_ds_cache[call_args] = ds
    return deepcopy(ds)


_create_bout_grid_ds_cache = {}


def create_bout_grid_ds(xsize=2, ysize=4, guards={}, topology="core", ny_inner=0):
    call_args = _get_kwargs()

    try:
        # Has been called with the same signature before, just return the cached result
        return deepcopy(_create_bout_grid_ds_cache[call_args])
    except KeyError:
        pass

    # Set the shape of the data in this dataset
    mxg = guards.get("x", 0)
    myg = guards.get("y", 0)
    xsize += 2 * mxg
    ysize += 2 * myg

    # jyseps* from grid file only ever used to check topology when loading the grid file,
    # so do not need to be consistent with the main dataset
    jyseps2_1 = ysize // 2
    jyseps1_2 = jyseps2_1

    if "double-null" in topology or "xpoint" in topology:
        # Has upper target as well
        ysize += 2 * myg

        # make different from jyseps2_1 so double-null toplogy is recognised
        jyseps1_2 += 1

    shape = (xsize, ysize)

    data = DataArray(np.ones(shape), dims=["x", "y"])

    ds = Dataset(
        {
            "psixy": data,
            "Rxy": data,
            "Zxy": data,
            "hthe": data,
            "y_boundary_guards": myg,
            "jyseps2_1": jyseps2_1,
            "jyseps1_2": jyseps1_2,
            "ny_inner": ny_inner,
        }
    )

    _create_bout_grid_ds_cache[call_args] = ds
    return deepcopy(ds)



import operator
import re
from copy import deepcopy
from functools import reduce
from pathlib import Path

import numpy as np
import pytest
from xarray import DataArray

from xbout.tests.utils_for_tests import (
    _get_kwargs,
    create_bout_ds_list,
    create_bout_grid_ds,
)


@pytest.fixture(scope="session")
def bout_xyt_example_files(tmp_path_factory):
    return _bout_xyt_example_files


_bout_xyt_example_files_cache = {}


def _bout_xyt_example_files(
    tmp_path_factory,
    prefix="BOUT.dmp",
    lengths=(6, 2, 4, 7),
    nxpe=4,
    nype=2,
    nt=1,
    guards=None,
    syn_data_type="random",
    grid=None,
    squashed=False,
    topology="core",
    write_to_disk=False,
    bout_v5=False,
    metric_3D=False,
):
    """
    Mocks up a set of BOUT-like Datasets

    Either returns list of Datasets (if write_to_disk=False)
    or writes Datasets to netCDF files and returns the temporary test directory
    containing them, deleting the temporary directory once that test is done (if
    write_to_disk=True).
    """
    call_args = _get_kwargs(ignore="tmp_path_factory")

    try:
        # Has been called with the same signature before, just return the cached result
        return deepcopy(_bout_xyt_example_files_cache[call_args])
    except KeyError:
        pass

    if guards is None:
        guards = {}

    if squashed:
        # create a single data-file, but alter the 'nxpe' and 'nype' variables, as if the
        # file had been created by combining a set of BOUT.dmp.*.nc files
        this_lengths = (
            lengths[0],
            lengths[1] * nxpe,
            lengths[2] * nype,
            lengths[3],
        )
        ds_list, file_list = create_bout_ds_list(
            prefix=prefix,
            lengths=this_lengths,
            nxpe=1,
            nype=1,
            nt=nt,
            guards=guards,
            topology=topology,
            syn_data_type=syn_data_type,
            squashed=True,
            bout_v5=bout_v5,
            metric_3D=metric_3D,
        )
        ds_list[0]["nxpe"] = nxpe
        ds_list[0]["nype"] = nype
    else:
        ds_list, file_list = create_bout_ds_list(
            prefix=prefix,
            lengths=lengths,
            nxpe=nxpe,
            nype=nype,
            nt=nt,
            guards=guards,
            topology=topology,
            syn_data_type=syn_data_type,
            bout_v5=bout_v5,
            metric_3D=metric_3D,
        )

    if grid is not None:
        xsize = lengths[1] * nxpe
        ysize = lengths[2] * nype
        grid_ds = create_bout_grid_ds(
            xsize=xsize,
            ysize=ysize,
            guards=guards,
            topology=topology,
            ny_inner=3 * lengths[2],
        )

    if not write_to_disk:
        if grid is None:
            _bout_xyt_example_files_cache[call_args] = ds_list
            return deepcopy(ds_list)
        else:
            _bout_xyt_example_files_cache[call_args] = ds_list, grid_ds
            return deepcopy((ds_list, grid_ds))
        raise ValueError("tmp_path_factory required when write_to_disk=True")

    save_dir = tmp_path_factory.mktemp("data")

    for ds, file_name in zip(ds_list, file_list):
        ds.to_netcdf(save_dir.joinpath(file_name))

    if grid is not None:
        grid_ds.to_netcdf(save_dir.joinpath(grid + ".nc"))

    # Return a glob-like path to all files created, which has all file numbers replaced
    # with a single asterix
    path = str(save_dir.joinpath(file_list[-1]))

    count = 1
    if nt > 1:
        count += 1
    # We have to reverse the path before limiting the number of numbers replaced so that the
    # tests don't get confused by pytest's persistent temporary directories (which are also designated
    # by different numbers)
    glob_pattern = Path((re.sub(r"\d+", "*", path[::-1], count=count))[::-1])
    _bout_xyt_example_files_cache[call_args] = glob_pattern
    return glob_pattern


fci_shape = (2, 2, 3, 4)
fci_guards = (2, 2, 0)


@pytest.fixture
def create_example_grid_file_fci(tmp_path_factory):
    """
    Mocks up a FCI-like netCDF file, and return the temporary test
    directory containing them.

    Deletes the temporary directory once that test is done.
    """

    # Create grid dataset
    shape = (fci_shape[1] + 2 * fci_guards[0], *fci_shape[2:])
    arr = np.arange(reduce(operator.mul, shape, 1)).reshape(shape)
    grid = DataArray(data=arr, name="R", dims=["x", "y", "z"]).to_dataset()
    grid["Z"] = DataArray(np.random.random(shape), dims=["x", "y", "z"])
    grid["dy"] = DataArray(np.ones(shape), dims=["x", "y", "z"])
    grid = grid.set_coords(["dy"])

    # Create temporary directory
    save_dir = tmp_path_factory.mktemp("griddata")

    # Save
    filepath = save_dir.joinpath("fci.nc")
    grid.to_netcdf(filepath, engine="netcdf4")

    return filepath


@pytest.fixture
def create_example_files_fci(tmp_path_factory):
    return _bout_xyt_example_files(
        tmp_path_factory,
        lengths=fci_shape,
        nxpe=1,
        nype=1,
        # nt=1,
        guards={a: b for a, b in zip("xyz", fci_guards)},
        syn_data_type="random",
        grid=None,
        squashed=False,
        # topology="core",
        write_to_disk=False,
        bout_v5=True,
        metric_3D=True,
    )



import pytest

from xarray import concat
from xbout import open_boutdataset


@pytest.mark.skip
class TestStormDataset:
    @pytest.mark.xfail
    def test_storm_dataset(self, tmpdir_factory, bout_xyt_example_files):
        path = bout_xyt_example_files(tmpdir_factory, nxpe=1, nype=1, nt=1)
        ds = open_boutdataset(datapath=path, inputfilepath=None)
        print(ds.storm.normalisation)

        assert False

    @pytest.mark.xfail
    def test_storm_dataset_inheritance(self, tmpdir_factory, bout_xyt_example_files):
        path = bout_xyt_example_files(tmpdir_factory, nxpe=1, nype=1, nt=1)
        ds = open_boutdataset(datapath=path, inputfilepath=None)
        ds.storm.set_extra_data("options")
        print(ds.storm.extra_data)

        assert False

    @pytest.mark.xfail
    def test_object_permanence(self, tmpdir_factory, bout_xyt_example_files):
        path = bout_xyt_example_files(tmpdir_factory, nxpe=1, nype=1, nt=1)
        ds = open_boutdataset(datapath=path, inputfilepath=None)

        ds.storm.extra_info = "options"
        new_ds = ds.isel(t=-1)
        print(new_ds.storm.extra_info)

        assert False

    @pytest.mark.xfail
    def test_dataset_duck_typing(self, tmpdir_factory, bout_xyt_example_files):
        path = bout_xyt_example_files(tmpdir_factory, nxpe=1, nype=1, nt=1)
        ds = open_boutdataset(datapath=path, inputfilepath=None)

        result = concat([ds.bout, ds.bout], dim="run")
        print(result)



from copy import deepcopy
import numpy as np
import xarray as xr
from .utils import _set_attrs_on_all_vars


class Region:
    """
    Contains the global indices bounding a single topological region, i.e. a region with
    logically rectangular contiguous data.

    Also stores the names of any neighbouring regions.
    """

    def __init__(
        self,
        *,
        name,
        ds=None,
        xinner_ind=None,
        xouter_ind=None,
        ylower_ind=None,
        yupper_ind=None,
        connection_inner_x=None,
        connection_outer_x=None,
        connection_lower_y=None,
        connection_upper_y=None,
    ):
        """
        Parameters
        ----------
        name : str
            Name of the region
        ds : BoutDataset, optional
            Dataset to get variables to calculate coordinates from
        xinner_ind : int, optional
            Global x-index of the inner points of this region
        xouter_ind : int, optional
            Global x-index of the points just beyond the outer edge of this region
        ylower_ind : int, optional
            Global y-index of the lower points of this region
        yupper_ind : int, optional
            Global y-index of the points just beyond the upper edge of this region
        connection_inner_x : str, optional
            The region inside this one in the x-direction
        connection_outer_x : str, optional
            The region outside this one in the x-direction
        connection_lower_y : str, optional
            The region below this one in the y-direction
        connection_upper_y : str, optional
            The region above this one in the y-direction
        """
        self.name = name
        self.xinner_ind = xinner_ind
        self.xouter_ind = xouter_ind
        if xouter_ind is not None and xinner_ind is not None:
            self.nx = xouter_ind - xinner_ind
        self.ylower_ind = ylower_ind
        self.yupper_ind = yupper_ind
        if yupper_ind is not None and ylower_ind is not None:
            self.ny = yupper_ind - ylower_ind
        self.connection_inner_x = connection_inner_x
        self.connection_outer_x = connection_outer_x
        self.connection_lower_y = connection_lower_y
        self.connection_upper_y = connection_upper_y
        self.global_nx = ds.sizes[ds.metadata["bout_xdim"]]
        self.global_ny = ds.sizes[ds.metadata["bout_ydim"]]

        if ds is not None:
            # self.nx, self.ny should not include boundary points.
            # self.xinner, self.xouter, self.ylower, self.yupper
            if ds.metadata["keep_xboundaries"]:
                xbndry = ds.metadata["MXG"]
                if self.connection_inner_x is None:
                    self.nx = self.nx - xbndry

                    # used to calculate x-coordinate of inner side (self.xinner)
                    xinner_ind = xinner_ind + xbndry

                if self.connection_outer_x is None:
                    self.nx = self.nx - xbndry

                    # used to calculate x-coordinate of outer side (self.xouter)
                    xouter_ind = xouter_ind - xbndry

            if ds.metadata["keep_yboundaries"]:
                ybndry = ds.metadata["MYG"]
                if self.connection_lower_y is None:
                    self.ny = self.ny - ybndry

                    # used to calculate y-coordinate of lower side (self.ylower)
                    ylower_ind = ylower_ind + ybndry

                if self.connection_upper_y is None:
                    self.ny = self.ny - ybndry

                    # used to calculate y-coordinate of upper side (self.yupper)
                    yupper_ind = yupper_ind - ybndry

            # calculate start and end coordinates
            #####################################
            self.xcoord = ds.metadata["bout_xdim"]
            self.ycoord = ds.metadata["bout_ydim"]

            # Note the global coordinates used here are defined so that they are zero at
            # the boundaries of the grid (where the grid includes all boundary cells),
            # not necessarily the physical boundaries because constant offsets do not
            # matter, as long as these bounds are consistent with the global coordinates
            # defined in apply_geometry (we will only use these coordinates for
            # interpolation) and it is simplest to calculate them with cumsum().

            # dx is constant in any particular region in the y-direction, so convert to a
            # 1d array
            # Note that this is not the same coordinate as the 'x' coordinate that is
            # created by default from the x-index, as these values are set only for
            # particular regions, so do not need to be consistent between different
            # regions (e.g. core and PFR), so we are not forced to use just the index
            # value here.
            # Define ref_yind so that we avoid using values from the corner cells, which
            # might be NaN.
            if self.connection_lower_y is None and ds.metadata["keep_yboundaries"] == 1:
                ref_yind = ylower_ind + ds.metadata["MYG"]
            else:
                ref_yind = ylower_ind
            dx = ds["dx"].isel({self.ycoord: ref_yind})
            dx_cumsum = dx.cumsum()
            self.xinner = dx_cumsum[xinner_ind] - dx[xinner_ind]
            self.xouter = dx_cumsum[xouter_ind - 1] + dx[xouter_ind - 1]

            # dy is constant in the x-direction, so convert to a 1d array
            # Define ref_xind so that we avoid using values from the corner cells, which
            # might be NaN.
            if self.connection_inner_x is None and ds.metadata["keep_xboundaries"] == 1:
                ref_xind = xinner_ind + ds.metadata["MXG"]
            else:
                ref_xind = xinner_ind
            dy = ds["dy"].isel(**{self.xcoord: ref_xind})
            dy_cumsum = dy.cumsum()
            self.ylower = dy_cumsum[ylower_ind] - dy[ylower_ind]
            self.yupper = dy_cumsum[yupper_ind - 1]

    def __repr__(self):
        result = "<xbout.region.Region>\n"
        for attr, val in vars(self).items():
            result += f"\t{attr}\t{val}\n"
        return result

    def get_slices(self, mxg=0, myg=0):
        """
        Return x- and y-dimension slices that select this region from the global
        DataArray.

        Returns
        -------
        xslice, yslice : slice, slice
        """
        xi = self.xinner_ind
        if self.connection_inner_x is not None and xi is not None:
            xi = xi - mxg

        xo = self.xouter_ind
        if self.connection_outer_x is not None and xo is not None:
            xo = xo + mxg

        yl = self.ylower_ind
        if self.connection_lower_y is not None and yl is not None:
            yl = yl - myg

        yu = self.yupper_ind
        if self.connection_upper_y is not None and yu is not None:
            yu = yu + myg

        return {self.xcoord: slice(xi, xo), self.ycoord: slice(yl, yu)}

    def get_inner_guards_slices(self, *, mxg, myg=0):
        """
        Return x- and y-dimension slices that select mxg guard cells on the inner-x side
        of this region from the global DataArray.

        Parameters
        ----------
        mxg : int
            Number of guard cells
        myg : int, optional
            Number of y-guard cells to include at the corners
        """
        ylower = self.ylower_ind
        if self.connection_lower_y is not None:
            ylower = ylower - myg
        yupper = self.yupper_ind
        if self.connection_upper_y is not None:
            yupper = yupper + myg
        return {
            self.xcoord: slice(self.xinner_ind - mxg, self.xinner_ind),
            self.ycoord: slice(ylower, yupper),
        }

    def get_outer_guards_slices(self, *, mxg, myg=0):
        """
        Return x- and y-dimension slices that select mxg guard cells on the outer-x side
        of this region from the global DataArray.

        Parameters
        ----------
        mxg : int
            Number of guard cells
        myg : int, optional
            Number of y-guard cells to include at the corners
        """
        ylower = self.ylower_ind
        if self.connection_lower_y is not None:
            ylower = ylower - myg
        yupper = self.yupper_ind
        if self.connection_upper_y is not None:
            yupper = yupper + myg
        return {
            self.xcoord: slice(self.xouter_ind, self.xouter_ind + mxg),
            self.ycoord: slice(ylower, yupper),
        }

    def get_lower_guards_slices(self, *, myg, mxg=0):
        """
        Return x- and y-dimension slices that select myg guard cells on the lower-y side
        of this region from the global DataArray.

        Parameters
        ----------
        myg : int
            Number of guard cells
        mxg : int, optional
            Number of x-guard cells to include at the corners
        """
        xinner = self.xinner_ind
        if self.connection_inner_x is not None:
            xinner = xinner - mxg
        xouter = self.xouter_ind
        if self.connection_outer_x is not None:
            xouter = xouter + mxg

        ystart = self.ylower_ind - myg
        ystop = self.ylower_ind
        if ystart < 0:
            # Make sure negative indices wrap around correctly
            if ystop == 0:
                ystop = None
            elif ystop > 0:
                raise ValueError(
                    f"'start={ystart}' of lower guards slice is negative, but "
                    f"'stop={ystop}' is positive. Should not be possible."
                )
        return {
            self.xcoord: slice(xinner, xouter),
            self.ycoord: slice(ystart, ystop),
        }

    def get_upper_guards_slices(self, *, myg, mxg=0):
        """
        Return x- and y-dimension slices that select myg guard cells on the upper-y side
        of this region from the global DataArray.

        Parameters
        ----------
        myg : int
            Number of guard cells
        mxg : int, optional
            Number of x-guard cells to include at the corners
        """
        xinner = self.xinner_ind
        if self.connection_inner_x is not None:
            xinner = xinner - mxg
        xouter = self.xouter_ind
        if self.connection_outer_x is not None:
            xouter = xouter + mxg

        # wrap around to the beginning of the grid if necessary
        ystart = self.yupper_ind
        ystop = self.yupper_ind + myg
        if ystart >= self.global_ny:
            # wrap around to beginning of global array
            ystart = ystart - self.global_ny
            ystop = ystop - self.global_ny
        return {
            self.xcoord: slice(xinner, xouter),
            self.ycoord: slice(ystart, ystop),
        }

    def __eq__(self, other):
        if not isinstance(other, Region):
            return NotImplemented
        return vars(self) == vars(other)


def _in_range(val, lower, upper):
    if val < lower:
        return lower
    elif val > upper:
        return upper
    else:
        return val


def _order_vars(lower, upper):
    if upper < lower:
        return upper, lower
    else:
        return lower, upper


def _get_topology(ds):
    jys11 = ds.metadata["jyseps1_1"]
    jys21 = ds.metadata["jyseps2_1"]
    ny_inner = ds.metadata["ny_inner"]
    jys12 = ds.metadata["jyseps1_2"]
    jys22 = ds.metadata["jyseps2_2"]
    ny = ds.metadata["ny"]
    ixs1 = ds.metadata["ixseps1"]
    ixs2 = ds.metadata["ixseps2"]
    nx = ds.metadata["nx"]
    if jys21 == jys12:
        # No upper X-point
        if jys11 <= 0 and jys22 >= ny - 1:
            ix = min(ixs1, ixs2)
            if ix >= nx - 1:
                return "core"
            elif ix <= 0:
                return "sol"
            else:
                return "limiter"

        return "single-null"

    if jys11 == jys21 and jys12 == jys22:
        if jys11 < ny_inner - 1 and jys22 > ny_inner:
            return "xpoint"
        else:
            raise ValueError("Currently unsupported topology")

    if ixs1 == ixs2:
        if jys21 < ny_inner - 1 and jys12 > ny_inner:
            return "connected-double-null"
        else:
            raise ValueError("Currently unsupported topology")
    elif ixs1 < ixs2:
        return "lower-disconnected-double-null"
    else:
        # ixs1 > ixs2
        return "upper-disconnected-double-null"


def _check_connections(regions):
    for region in regions.values():
        if region.connection_inner_x is not None:
            if regions[region.connection_inner_x].connection_outer_x != region.name:
                raise ValueError(
                    f"Inner-x connection of {region.name} is "
                    f"{region.connection_inner_x}, but outer-x connection of "
                    f"{region.connection_inner_x} is "
                    f"{regions[region.connection_inner_x].connection_outer_x}"
                )
        if region.connection_outer_x is not None:
            if regions[region.connection_outer_x].connection_inner_x != region.name:
                raise ValueError(
                    f"Inner-x connection of {region.name} is "
                    f"{region.connection_outer_x}, but inner-x connection of "
                    f"{region.connection_outer_x} is "
                    f"{regions[region.connection_outer_x].connection_inner_x}"
                )
        if region.connection_lower_y is not None:
            if regions[region.connection_lower_y].connection_upper_y != region.name:
                raise ValueError(
                    f"Lower-y connection of {region.name} is "
                    f"{region.connection_lower_y}, but upper-y connection of "
                    f"{region.connection_lower_y} is "
                    f"{regions[region.connection_lower_y].connection_upper_y}"
                )
        if region.connection_upper_y is not None:
            if regions[region.connection_upper_y].connection_lower_y != region.name:
                raise ValueError(
                    f"Upper-y connection of {region.name} is "
                    f"{region.connection_upper_y}, but lower-y connection of "
                    f"{region.connection_upper_y} is "
                    f"{regions[region.connection_upper_y].connection_lower_y}"
                )


topologies = {}


def topology_lower_disconnected_double_null(
    *, ds, ixs1, ixs2, nx, jys11, jys21, ny_inner, jys12, jys22, ny, ybndry
):
    regions = {}
    regions["lower_inner_PFR"] = Region(
        name="lower_inner_PFR",
        ds=ds,
        xinner_ind=0,
        xouter_ind=ixs1,
        ylower_ind=0,
        yupper_ind=jys11 + 1,
        connection_outer_x="lower_inner_intersep",
        connection_upper_y="lower_outer_PFR",
    )
    regions["lower_inner_intersep"] = Region(
        name="lower_inner_intersep",
        ds=ds,
        xinner_ind=ixs1,
        xouter_ind=ixs2,
        ylower_ind=0,
        yupper_ind=jys11 + 1,
        connection_inner_x="lower_inner_PFR",
        connection_outer_x="lower_inner_SOL",
        connection_upper_y="inner_intersep",
    )
    regions["lower_inner_SOL"] = Region(
        name="lower_inner_SOL",
        ds=ds,
        xinner_ind=ixs2,
        xouter_ind=nx,
        ylower_ind=0,
        yupper_ind=jys11 + 1,
        connection_inner_x="lower_inner_intersep",
        connection_upper_y="inner_SOL",
    )
    regions["inner_core"] = Region(
        name="inner_core",
        ds=ds,
        xinner_ind=0,
        xouter_ind=ixs1,
        ylower_ind=jys11 + 1,
        yupper_ind=jys21 + 1,
        connection_outer_x="inner_intersep",
        connection_lower_y="outer_core",
        connection_upper_y="outer_core",
    )
    regions["inner_intersep"] = Region(
        name="inner_intersep",
        ds=ds,
        xinner_ind=ixs1,
        xouter_ind=ixs2,
        ylower_ind=jys11 + 1,
        yupper_ind=jys21 + 1,
        connection_inner_x="inner_core",
        connection_outer_x="inner_SOL",
        connection_lower_y="lower_inner_intersep",
        connection_upper_y="outer_intersep",
    )
    regions["inner_SOL"] = Region(
        name="inner_SOL",
        ds=ds,
        xinner_ind=ixs2,
        xouter_ind=nx,
        ylower_ind=jys11 + 1,
        yupper_ind=jys21 + 1,
        connection_inner_x="inner_intersep",
        connection_lower_y="lower_inner_SOL",
        connection_upper_y="upper_inner_SOL",
    )
    regions["upper_inner_PFR"] = Region(
        name="upper_inner_PFR",
        ds=ds,
        xinner_ind=0,
        xouter_ind=ixs1,
        ylower_ind=jys21 + 1,
        yupper_ind=ny_inner,
        connection_outer_x="upper_inner_intersep",
        connection_lower_y="upper_outer_PFR",
    )
    regions["upper_inner_intersep"] = Region(
        name="upper_inner_intersep",
        ds=ds,
        xinner_ind=ixs1,
        xouter_ind=ixs2,
        ylower_ind=jys21 + 1,
        yupper_ind=ny_inner,
        connection_inner_x="upper_inner_PFR",
        connection_outer_x="upper_inner_SOL",
        connection_lower_y="upper_outer_intersep",
    )
    regions["upper_inner_SOL"] = Region(
        name="upper_inner_SOL",
        ds=ds,
        xinner_ind=ixs2,
        xouter_ind=nx,
        ylower_ind=jys21 + 1,
        yupper_ind=ny_inner,
        connection_inner_x="upper_inner_intersep",
        connection_lower_y="inner_SOL",
    )
    regions["upper_outer_PFR"] = Region(
        name="upper_outer_PFR",
        ds=ds,
        xinner_ind=0,
        xouter_ind=ixs1,
        ylower_ind=ny_inner,
        yupper_ind=jys12 + 1,
        connection_outer_x="upper_outer_intersep",
        connection_upper_y="upper_inner_PFR",
    )
    regions["upper_outer_intersep"] = Region(
        name="upper_outer_intersep",
        ds=ds,
        xinner_ind=ixs1,
        xouter_ind=ixs2,
        ylower_ind=ny_inner,
        yupper_ind=jys12 + 1,
        connection_inner_x="upper_outer_PFR",
        connection_outer_x="upper_outer_SOL",
        connection_upper_y="upper_inner_intersep",
    )
    regions["upper_outer_SOL"] = Region(
        name="upper_outer_SOL",
        ds=ds,
        xinner_ind=ixs2,
        xouter_ind=nx,
        ylower_ind=ny_inner,
        yupper_ind=jys12 + 1,
        connection_inner_x="upper_outer_intersep",
        connection_upper_y="outer_SOL",
    )
    regions["outer_core"] = Region(
        name="outer_core",
        ds=ds,
        xinner_ind=0,
        xouter_ind=ixs1,
        ylower_ind=jys12 + 1,
        yupper_ind=jys22 + 1,
        connection_outer_x="outer_intersep",
        connection_lower_y="inner_core",
        connection_upper_y="inner_core",
    )
    regions["outer_intersep"] = Region(
        name="outer_intersep",
        ds=ds,
        xinner_ind=ixs1,
        xouter_ind=ixs2,
        ylower_ind=jys12 + 1,
        yupper_ind=jys22 + 1,
        connection_inner_x="outer_core",
        connection_outer_x="outer_SOL",
        connection_lower_y="inner_intersep",
        connection_upper_y="lower_outer_intersep",
    )
    regions["outer_SOL"] = Region(
        name="outer_SOL",
        ds=ds,
        xinner_ind=ixs2,
        xouter_ind=nx,
        ylower_ind=jys12 + 1,
        yupper_ind=jys22 + 1,
        connection_inner_x="outer_intersep",
        connection_lower_y="upper_outer_SOL",
        connection_upper_y="lower_outer_SOL",
    )
    regions["lower_outer_PFR"] = Region(
        name="lower_outer_PFR",
        ds=ds,
        xinner_ind=0,
        xouter_ind=ixs1,
        ylower_ind=jys22 + 1,
        yupper_ind=ny,
        connection_outer_x="lower_outer_intersep",
        connection_lower_y="lower_inner_PFR",
    )
    regions["lower_outer_intersep"] = Region(
        name="lower_outer_intersep",
        ds=ds,
        xinner_ind=ixs1,
        xouter_ind=ixs2,
        ylower_ind=jys22 + 1,
        yupper_ind=ny,
        connection_inner_x="lower_outer_PFR",
        connection_outer_x="lower_outer_SOL",
        connection_lower_y="outer_intersep",
    )
    regions["lower_outer_SOL"] = Region(
        name="lower_outer_SOL",
        ds=ds,
        xinner_ind=ixs2,
        xouter_ind=nx,
        ylower_ind=jys22 + 1,
        yupper_ind=ny,
        connection_inner_x="lower_outer_intersep",
        connection_lower_y="outer_SOL",
    )
    return regions


topologies["lower-disconnected-double-null"] = topology_lower_disconnected_double_null


def topology_upper_disconnected_double_null(
    *, ds, ixs1, ixs2, nx, jys11, jys21, ny_inner, jys12, jys22, ny, ybndry
):
    regions = {}
    regions["lower_inner_PFR"] = Region(
        name="lower_inner_PFR",
        ds=ds,
        xinner_ind=0,
        xouter_ind=ixs2,
        ylower_ind=0,
        yupper_ind=jys11 + 1,
        connection_outer_x="lower_inner_intersep",
        connection_upper_y="lower_outer_PFR",
    )
    regions["lower_inner_intersep"] = Region(
        name="lower_inner_intersep",
        ds=ds,
        xinner_ind=ixs2,
        xouter_ind=ixs1,
        ylower_ind=0,
        yupper_ind=jys11 + 1,
        connection_inner_x="lower_inner_PFR",
        connection_outer_x="lower_inner_SOL",
        connection_upper_y="lower_outer_intersep",
    )
    regions["lower_inner_SOL"] = Region(
        name="lower_inner_SOL",
        ds=ds,
        xinner_ind=ixs1,
        xouter_ind=nx,
        ylower_ind=0,
        yupper_ind=jys11 + 1,
        connection_inner_x="lower_inner_intersep",
        connection_upper_y="inner_SOL",
    )
    regions["inner_core"] = Region(
        name="inner_core",
        ds=ds,
        xinner_ind=0,
        xouter_ind=ixs2,
        ylower_ind=jys11 + 1,
        yupper_ind=jys21 + 1,
        connection_outer_x="inner_intersep",
        connection_lower_y="outer_core",
        connection_upper_y="outer_core",
    )
    regions["inner_intersep"] = Region(
        name="inner_intersep",
        ds=ds,
        xinner_ind=ixs2,
        xouter_ind=ixs1,
        ylower_ind=jys11 + 1,
        yupper_ind=jys21 + 1,
        connection_inner_x="inner_core",
        connection_outer_x="inner_SOL",
        connection_lower_y="outer_intersep",
        connection_upper_y="upper_inner_intersep",
    )
    regions["inner_SOL"] = Region(
        name="inner_SOL",
        ds=ds,
        xinner_ind=ixs1,
        xouter_ind=nx,
        ylower_ind=jys11 + 1,
        yupper_ind=jys21 + 1,
        connection_inner_x="inner_intersep",
        connection_lower_y="lower_inner_SOL",
        connection_upper_y="upper_inner_SOL",
    )
    regions["upper_inner_PFR"] = Region(
        name="upper_inner_PFR",
        ds=ds,
        xinner_ind=0,
        xouter_ind=ixs2,
        ylower_ind=jys21 + 1,
        yupper_ind=ny_inner,
        connection_outer_x="upper_inner_intersep",
        connection_lower_y="upper_outer_PFR",
    )
    regions["upper_inner_intersep"] = Region(
        name="upper_inner_intersep",
        ds=ds,
        xinner_ind=ixs2,
        xouter_ind=ixs1,
        ylower_ind=jys21 + 1,
        yupper_ind=ny_inner,
        connection_inner_x="upper_inner_PFR",
        connection_outer_x="upper_inner_SOL",
        connection_lower_y="inner_intersep",
    )
    regions["upper_inner_SOL"] = Region(
        name="upper_inner_SOL",
        ds=ds,
        xinner_ind=ixs1,
        xouter_ind=nx,
        ylower_ind=jys21 + 1,
        yupper_ind=ny_inner,
        connection_inner_x="upper_inner_intersep",
        connection_lower_y="inner_SOL",
    )
    regions["upper_outer_PFR"] = Region(
        name="upper_outer_PFR",
        ds=ds,
        xinner_ind=0,
        xouter_ind=ixs2,
        ylower_ind=ny_inner,
        yupper_ind=jys12 + 1,
        connection_outer_x="upper_outer_intersep",
        connection_upper_y="upper_inner_PFR",
    )
    regions["upper_outer_intersep"] = Region(
        name="upper_outer_intersep",
        ds=ds,
        xinner_ind=ixs2,
        xouter_ind=ixs1,
        ylower_ind=ny_inner,
        yupper_ind=jys12 + 1,
        connection_inner_x="upper_outer_PFR",
        connection_outer_x="upper_outer_SOL",
        connection_upper_y="outer_intersep",
    )
    regions["upper_outer_SOL"] = Region(
        name="upper_outer_SOL",
        ds=ds,
        xinner_ind=ixs1,
        xouter_ind=nx,
        ylower_ind=ny_inner,
        yupper_ind=jys12 + 1,
        connection_inner_x="upper_outer_intersep",
        connection_upper_y="outer_SOL",
    )
    regions["outer_core"] = Region(
        name="outer_core",
        ds=ds,
        xinner_ind=0,
        xouter_ind=ixs2,
        ylower_ind=jys12 + 1,
        yupper_ind=jys22 + 1,
        connection_outer_x="outer_intersep",
        connection_lower_y="inner_core",
        connection_upper_y="inner_core",
    )
    regions["outer_intersep"] = Region(
        name="outer_intersep",
        ds=ds,
        xinner_ind=ixs2,
        xouter_ind=ixs1,
        ylower_ind=jys12 + 1,
        yupper_ind=jys22 + 1,
        connection_inner_x="outer_core",
        connection_outer_x="outer_SOL",
        connection_lower_y="upper_outer_intersep",
        connection_upper_y="inner_intersep",
    )
    regions["outer_SOL"] = Region(
        name="outer_SOL",
        ds=ds,
        xinner_ind=ixs1,
        xouter_ind=nx,
        ylower_ind=jys12 + 1,
        yupper_ind=jys22 + 1,
        connection_inner_x="outer_intersep",
        connection_lower_y="upper_outer_SOL",
        connection_upper_y="lower_outer_SOL",
    )
    regions["lower_outer_PFR"] = Region(
        name="lower_outer_PFR",
        ds=ds,
        xinner_ind=0,
        xouter_ind=ixs2,
        ylower_ind=jys22 + 1,
        yupper_ind=ny,
        connection_outer_x="lower_outer_intersep",
        connection_lower_y="lower_inner_PFR",
    )
    regions["lower_outer_intersep"] = Region(
        name="lower_outer_intersep",
        ds=ds,
        xinner_ind=ixs2,
        xouter_ind=ixs1,
        ylower_ind=jys22 + 1,
        yupper_ind=ny,
        connection_inner_x="lower_outer_PFR",
        connection_outer_x="lower_outer_SOL",
        connection_lower_y="lower_inner_intersep",
    )
    regions["lower_outer_SOL"] = Region(
        name="lower_outer_SOL",
        ds=ds,
        xinner_ind=ixs1,
        xouter_ind=nx,
        ylower_ind=jys22 + 1,
        yupper_ind=ny,
        connection_inner_x="lower_outer_intersep",
        connection_lower_y="outer_SOL",
    )
    return regions


topologies["upper-disconnected-double-null"] = topology_upper_disconnected_double_null


def topology_connected_double_null(
    *, ds, ixs1, ixs2, nx, jys11, jys21, ny_inner, jys12, jys22, ny, ybndry
):
    regions = {}
    regions["lower_inner_PFR"] = Region(
        name="lower_inner_PFR",
        ds=ds,
        xinner_ind=0,
        xouter_ind=ixs1,
        ylower_ind=0,
        yupper_ind=jys11 + 1,
        connection_outer_x="lower_inner_SOL",
        connection_upper_y="lower_outer_PFR",
    )
    regions["lower_inner_SOL"] = Region(
        name="lower_inner_SOL",
        ds=ds,
        xinner_ind=ixs2,
        xouter_ind=nx,
        ylower_ind=0,
        yupper_ind=jys11 + 1,
        connection_inner_x="lower_inner_PFR",
        connection_upper_y="inner_SOL",
    )
    regions["inner_core"] = Region(
        name="inner_core",
        ds=ds,
        xinner_ind=0,
        xouter_ind=ixs1,
        ylower_ind=jys11 + 1,
        yupper_ind=jys21 + 1,
        connection_outer_x="inner_SOL",
        connection_lower_y="outer_core",
        connection_upper_y="outer_core",
    )
    regions["inner_SOL"] = Region(
        name="inner_SOL",
        ds=ds,
        xinner_ind=ixs2,
        xouter_ind=nx,
        ylower_ind=jys11 + 1,
        yupper_ind=jys21 + 1,
        connection_inner_x="inner_core",
        connection_lower_y="lower_inner_SOL",
        connection_upper_y="upper_inner_SOL",
    )
    regions["upper_inner_PFR"] = Region(
        name="upper_inner_PFR",
        ds=ds,
        xinner_ind=0,
        xouter_ind=ixs1,
        ylower_ind=jys21 + 1,
        yupper_ind=ny_inner,
        connection_outer_x="upper_inner_SOL",
        connection_lower_y="upper_outer_PFR",
    )
    regions["upper_inner_SOL"] = Region(
        name="upper_inner_SOL",
        ds=ds,
        xinner_ind=ixs2,
        xouter_ind=nx,
        ylower_ind=jys21 + 1,
        yupper_ind=ny_inner,
        connection_inner_x="upper_inner_PFR",
        connection_lower_y="inner_SOL",
    )
    regions["upper_outer_PFR"] = Region(
        name="upper_outer_PFR",
        ds=ds,
        xinner_ind=0,
        xouter_ind=ixs1,
        ylower_ind=ny_inner,
        yupper_ind=jys12 + 1,
        connection_outer_x="upper_outer_SOL",
        connection_upper_y="upper_inner_PFR",
    )
    regions["upper_outer_SOL"] = Region(
        name="upper_outer_SOL",
        ds=ds,
        xinner_ind=ixs2,
        xouter_ind=nx,
        ylower_ind=ny_inner,
        yupper_ind=jys12 + 1,
        connection_inner_x="upper_outer_PFR",
        connection_upper_y="outer_SOL",
    )
    regions["outer_core"] = Region(
        name="outer_core",
        ds=ds,
        xinner_ind=0,
        xouter_ind=ixs1,
        ylower_ind=jys12 + 1,
        yupper_ind=jys22 + 1,
        connection_outer_x="outer_SOL",
        connection_lower_y="inner_core",
        connection_upper_y="inner_core",
    )
    regions["outer_SOL"] = Region(
        name="outer_SOL",
        ds=ds,
        xinner_ind=ixs2,
        xouter_ind=nx,
        ylower_ind=jys12 + 1,
        yupper_ind=jys22 + 1,
        connection_inner_x="outer_core",
        connection_lower_y="upper_outer_SOL",
        connection_upper_y="lower_outer_SOL",
    )
    regions["lower_outer_PFR"] = Region(
        name="lower_outer_PFR",
        ds=ds,
        xinner_ind=0,
        xouter_ind=ixs1,
        ylower_ind=jys22 + 1,
        yupper_ind=ny,
        connection_outer_x="lower_outer_SOL",
        connection_lower_y="lower_inner_PFR",
    )
    regions["lower_outer_SOL"] = Region(
        name="lower_outer_SOL",
        ds=ds,
        xinner_ind=ixs2,
        xouter_ind=nx,
        ylower_ind=jys22 + 1,
        yupper_ind=ny,
        connection_inner_x="lower_outer_PFR",
        connection_lower_y="outer_SOL",
    )
    return regions


topologies["connected-double-null"] = topology_connected_double_null


def topology_single_null(
    *, ds, ixs1, ixs2, nx, jys11, jys21, ny_inner, jys12, jys22, ny, ybndry
):
    regions = {}
    regions["inner_PFR"] = Region(
        name="inner_PFR",
        ds=ds,
        xinner_ind=0,
        xouter_ind=ixs1,
        ylower_ind=0,
        yupper_ind=jys11 + 1,
        connection_outer_x="inner_SOL",
        connection_upper_y="outer_PFR",
    )
    regions["inner_SOL"] = Region(
        name="inner_SOL",
        ds=ds,
        xinner_ind=ixs1,
        xouter_ind=nx,
        ylower_ind=0,
        yupper_ind=jys11 + 1,
        connection_inner_x="inner_PFR",
        connection_upper_y="SOL",
    )
    regions["core"] = Region(
        name="core",
        ds=ds,
        xinner_ind=0,
        xouter_ind=ixs1,
        ylower_ind=jys11 + 1,
        yupper_ind=jys22 + 1,
        connection_outer_x="SOL",
        connection_lower_y="core",
        connection_upper_y="core",
    )
    regions["SOL"] = Region(
        name="SOL",
        ds=ds,
        xinner_ind=ixs1,
        xouter_ind=nx,
        ylower_ind=jys11 + 1,
        yupper_ind=jys22 + 1,
        connection_inner_x="core",
        connection_lower_y="inner_SOL",
        connection_upper_y="outer_SOL",
    )
    regions["outer_PFR"] = Region(
        name="outer_PFR",
        ds=ds,
        xinner_ind=0,
        xouter_ind=ixs1,
        ylower_ind=jys22 + 1,
        yupper_ind=ny,
        connection_outer_x="outer_SOL",
        connection_lower_y="inner_PFR",
    )
    regions["outer_SOL"] = Region(
        name="outer_SOL",
        ds=ds,
        xinner_ind=ixs1,
        xouter_ind=nx,
        ylower_ind=jys22 + 1,
        yupper_ind=ny,
        connection_inner_x="outer_PFR",
        connection_lower_y="SOL",
    )
    return regions


topologies["single-null"] = topology_single_null


def topology_limiter(
    *, ds, ixs1, ixs2, nx, jys11, jys21, ny_inner, jys12, jys22, ny, ybndry
):
    regions = {}
    regions["core"] = Region(
        name="core",
        ds=ds,
        xinner_ind=0,
        xouter_ind=ixs1,
        ylower_ind=ybndry,
        yupper_ind=ny - ybndry,
        connection_outer_x="SOL",
        connection_lower_y="core",
        connection_upper_y="core",
    )
    regions["SOL"] = Region(
        name="SOL",
        ds=ds,
        xinner_ind=ixs1,
        xouter_ind=nx,
        ylower_ind=0,
        yupper_ind=ny,
        connection_inner_x="core",
    )
    return regions


topologies["limiter"] = topology_limiter


def topology_core(
    *, ds, ixs1, ixs2, nx, jys11, jys21, ny_inner, jys12, jys22, ny, ybndry
):
    regions = {}
    regions["core"] = Region(
        name="core",
        ds=ds,
        xinner_ind=0,
        xouter_ind=nx,
        ylower_ind=ybndry,
        yupper_ind=ny - ybndry,
        connection_lower_y="core",
        connection_upper_y="core",
    )
    return regions


topologies["core"] = topology_core


def topology_sol(
    *, ds, ixs1, ixs2, nx, jys11, jys21, ny_inner, jys12, jys22, ny, ybndry
):
    regions = {}
    regions["SOL"] = Region(
        name="SOL", ds=ds, xinner_ind=0, xouter_ind=nx, ylower_ind=0, yupper_ind=ny
    )
    return regions


topologies["sol"] = topology_sol


def topology_xpoint(
    *, ds, ixs1, ixs2, nx, jys11, jys21, ny_inner, jys12, jys22, ny, ybndry
):
    regions = {}
    regions["lower_inner_PFR"] = Region(
        name="lower_inner_PFR",
        ds=ds,
        xinner_ind=0,
        xouter_ind=ixs1,
        ylower_ind=0,
        yupper_ind=jys11 + 1,
        connection_outer_x="lower_inner_SOL",
        connection_upper_y="lower_outer_PFR",
    )
    regions["lower_inner_SOL"] = Region(
        name="lower_inner_SOL",
        ds=ds,
        xinner_ind=ixs1,
        xouter_ind=nx,
        ylower_ind=0,
        yupper_ind=jys11 + 1,
        connection_inner_x="lower_inner_PFR",
        connection_upper_y="upper_inner_SOL",
    )
    regions["upper_inner_PFR"] = Region(
        name="upper_inner_PFR",
        ds=ds,
        xinner_ind=0,
        xouter_ind=ixs1,
        ylower_ind=jys11 + 1,
        yupper_ind=ny_inner,
        connection_outer_x="upper_inner_SOL",
        connection_lower_y="upper_outer_PFR",
    )
    regions["upper_inner_SOL"] = Region(
        name="upper_inner_SOL",
        ds=ds,
        xinner_ind=ixs1,
        xouter_ind=nx,
        ylower_ind=jys11 + 1,
        yupper_ind=ny_inner,
        connection_inner_x="upper_inner_PFR",
        connection_lower_y="lower_inner_SOL",
    )
    regions["upper_outer_PFR"] = Region(
        name="upper_outer_PFR",
        ds=ds,
        xinner_ind=0,
        xouter_ind=ixs1,
        ylower_ind=ny_inner,
        yupper_ind=jys22 + 1,
        connection_outer_x="upper_outer_SOL",
        connection_upper_y="upper_inner_PFR",
    )
    regions["upper_outer_SOL"] = Region(
        name="upper_outer_SOL",
        ds=ds,
        xinner_ind=ixs1,
        xouter_ind=nx,
        ylower_ind=ny_inner,
        yupper_ind=jys22 + 1,
        connection_inner_x="upper_outer_PFR",
        connection_upper_y="lower_outer_SOL",
    )
    regions["lower_outer_PFR"] = Region(
        name="lower_outer_PFR",
        ds=ds,
        xinner_ind=0,
        xouter_ind=ixs1,
        ylower_ind=jys22 + 1,
        yupper_ind=ny,
        connection_outer_x="lower_outer_SOL",
        connection_lower_y="lower_inner_PFR",
    )
    regions["lower_outer_SOL"] = Region(
        name="lower_outer_SOL",
        ds=ds,
        xinner_ind=ixs1,
        xouter_ind=nx,
        ylower_ind=jys22 + 1,
        yupper_ind=ny,
        connection_inner_x="lower_outer_PFR",
        connection_lower_y="upper_outer_SOL",
    )
    return regions


topologies["xpoint"] = topology_xpoint


def _create_regions_toroidal(ds):
    topology = _get_topology(ds)

    ixs1 = ds.metadata["ixseps1"]
    ixs2 = ds.metadata["ixseps2"]
    nx = ds.metadata["nx"]

    jys11 = ds.metadata["jyseps1_1"]
    jys21 = ds.metadata["jyseps2_1"]
    ny_inner = ds.metadata["ny_inner"]
    jys12 = ds.metadata["jyseps1_2"]
    jys22 = ds.metadata["jyseps2_2"]
    ny = ds.metadata["ny"]

    mxg = ds.metadata["MXG"]
    myg = ds.metadata["MYG"]
    # keep_yboundaries is 1 if there are y-boundaries and 0 if there are not
    ybndry = ds.metadata["keep_yboundaries"] * myg
    if jys21 == jys12:
        # No upper targets
        ybndry_upper = 0
    else:
        ybndry_upper = ybndry

    # Make sure all sizes are sensible
    ixs1 = _in_range(ixs1, 0, nx)
    ixs2 = _in_range(ixs2, 0, nx)
    jys11 = _in_range(jys11, 0, ny - 1)
    jys21 = _in_range(jys21, 0, ny - 1)
    jys12 = _in_range(jys12, 0, ny - 1)
    jys21, jys12 = _order_vars(jys21, jys12)
    ny_inner = _in_range(ny_inner, jys21 + 1, jys12 + 1)
    jys22 = _in_range(jys22, 0, ny - 1)

    # Adjust for boundary cells
    # keep_xboundaries is 1 if there are x-boundaries and 0 if there are not
    if not ds.metadata["keep_xboundaries"]:
        ixs1 = ixs1 - mxg
        ixs2 = ixs2 - mxg
        nx = nx - 2 * mxg
    jys11 = jys11 + ybndry
    jys21 = jys21 + ybndry
    ny_inner = ny_inner + ybndry + ybndry_upper
    jys12 = jys12 + ybndry + 2 * ybndry_upper
    jys22 = jys22 + ybndry + 2 * ybndry_upper
    ny = ny + 2 * ybndry + 2 * ybndry_upper

    # Note, include guard cells in the created regions, fill them later
    if topology in topologies:
        regions = topologies[topology](
            ds=ds,
            ixs1=ixs1,
            ixs2=ixs2,
            nx=nx,
            jys11=jys11,
            jys21=jys21,
            ny_inner=ny_inner,
            jys12=jys12,
            jys22=jys22,
            ny=ny,
            ybndry=ybndry,
        )
    else:
        raise NotImplementedError(f"Topology '{topology}' is not implemented")

    _check_connections(regions)

    ds = _set_attrs_on_all_vars(ds, "regions", regions)
    ds.metadata["topology"] = topology

    return ds


def _create_single_region(ds, periodic_y=True):
    nx = ds.metadata["nx"]
    ny = ds.metadata["ny"]

    connection = "all" if periodic_y else None

    regions = {
        "all": Region(
            name="all",
            ds=ds,
            xouter_ind=nx,
            xinner_ind=0,
            ylower_ind=0,
            yupper_ind=ny,
            connection_lower_y=connection,
            connection_upper_y=connection,
        )
    }

    _check_connections(regions)

    ds = _set_attrs_on_all_vars(ds, "regions", regions)

    return ds


def _concat_inner_guards(da, da_global, mxg):
    """
    Concatenate inner x-guard cells to da, which is in a single region, getting the guard
    cells from the global array
    """

    if mxg <= 0:
        return da

    if len(da.bout._regions) > 1:
        raise ValueError("da passed should have only one region")
    region = list(da.bout._regions.values())[0]

    if region.connection_inner_x not in da_global.bout._regions:
        # No connection, or plotting restricted set of regions not including this
        # connection
        return da

    myg_da = da_global.metadata["MYG"]
    keep_yboundaries = da_global.metadata["keep_yboundaries"]
    xcoord = da_global.metadata["bout_xdim"]
    ycoord = da_global.metadata["bout_ydim"]

    da_inner = da_global.bout.from_region(region.connection_inner_x, with_guards=0)
    if len(da_inner.bout._regions) > 1:
        raise ValueError("da_inner should have only one region")
    region_inner = list(da_inner.bout._regions.values())[0]

    if (
        myg_da > 0
        and keep_yboundaries
        and region.connection_lower_y is not None
        and region_inner.connection_lower_y is None
    ):
        # da_inner may have more points in the y-direction, because it has an actual
        # boundary, not a connection. Need to remove any extra points
        da_inner = da_inner.isel(**{ycoord: slice(myg_da, None)})
    if (
        myg_da > 0
        and keep_yboundaries
        and region.connection_upper_y is not None
        and region_inner.connection_upper_y is None
    ):
        # da_inner may have more points in the y-direction, because it has an actual
        # boundary, not a connection. Need to remove any extra points
        da_inner = da_inner.isel(**{ycoord: slice(None, -myg_da)})

    # select just the points we need to fill the guard cells of da
    da_inner = da_inner.isel(**{xcoord: slice(-mxg, None)})

    if (
        myg_da > 0
        and keep_yboundaries
        and region.connection_lower_y is None
        and region_inner.connection_lower_y is not None
    ):
        # da_inner may have fewer points in the y-direction, because it has a connection,
        # not an actual boundary. Need to get the extra points from its connection
        da_inner_lower = da_global.bout.from_region(
            region_inner.connection_lower_y, with_guards=0
        )
        da_inner_lower = da_inner_lower.isel(
            **{xcoord: slice(-mxg, None), ycoord: slice(-myg_da, None)}
        )
        save_regions = da_inner.bout._regions
        da_inner = xr.concat((da_inner_lower, da_inner), ycoord, join="exact")
        # xr.concat takes attributes from the first variable, but we need da_inner's
        # regions
        da_inner.attrs["regions"] = save_regions
    if (
        myg_da > 0
        and keep_yboundaries
        and region.connection_upper_y is None
        and region_inner.connection_upper_y is not None
    ):
        # da_inner may have fewer points in the y-direction, because it has a connection,
        # not an actual boundary. Need to get the extra points from its connection
        da_inner_upper = da_global.bout.from_region(
            region_inner.connection_upper_y, with_guards=0
        )
        da_inner_upper = da_inner_upper.isel(
            **{xcoord: slice(-mxg, None), ycoord: slice(myg_da)}
        )
        da_inner = xr.concat((da_inner, da_inner_upper), ycoord, join="exact")
        # xr.concat takes attributes from the first variable, so regions are OK

    if xcoord in da.coords:
        # Some coordinates may not be single-valued, so use local coordinates for
        # neighbouring region, not communicated ones. Ensures the coordinates are
        # continuous so that interpolation works correctly near the boundaries.
        slices = region.get_inner_guards_slices(mxg=mxg)
        new_xcoord = da_global[xcoord].isel(**{xcoord: slices[xcoord]})
        new_ycoord = da_global[ycoord].isel(**{ycoord: slices[ycoord]})

        # can't use commented out version, uncommented one works around xarray bug
        # removing attrs
        # https://github.com/pydata/xarray/issues/4415
        # https://github.com/pydata/xarray/issues/4393
        # da_inner = da_inner.assign_coords(**{xcoord: new_xcoord, ycoord: new_ycoord})
        da_inner[xcoord].data[...] = new_xcoord.data
        da_inner[ycoord].data[...] = new_ycoord.data

    save_regions = da.bout._regions
    da = xr.concat((da_inner, da), xcoord, join="exact")
    # xr.concat takes attributes from the first variable (for xarray>=0.15.0, keeps attrs
    # that are the same in all objects for xarray<0.15.0)
    da.attrs["regions"] = save_regions

    return da


def _concat_outer_guards(da, da_global, mxg):
    """
    Concatenate outer x-guard cells to da, which is in a single region, getting the guard
    cells from the global array
    """

    if mxg <= 0:
        return da

    if len(da.bout._regions) > 1:
        raise ValueError("da passed should have only one region")
    region = list(da.bout._regions.values())[0]

    if region.connection_outer_x not in da_global.bout._regions:
        # No connection, or plotting restricted set of regions not including this
        # connection
        return da

    myg_da = da_global.metadata["MYG"]
    keep_yboundaries = da_global.metadata["keep_yboundaries"]
    xcoord = da_global.metadata["bout_xdim"]
    ycoord = da_global.metadata["bout_ydim"]

    da_outer = da_global.bout.from_region(region.connection_outer_x, with_guards=0)
    if len(da_outer.bout._regions) > 1:
        raise ValueError("da_outer should have only one region")
    region_outer = list(da_outer.bout._regions.values())[0]

    if (
        myg_da > 0
        and keep_yboundaries
        and region.connection_lower_y is not None
        and region_outer.connection_lower_y is None
    ):
        # da_outer may have more points in the y-direction, because it has an actual
        # boundary, not a connection. Need to remove any extra points
        da_outer = da_outer.isel(**{ycoord: slice(myg_da, None)})
    if (
        myg_da > 0
        and keep_yboundaries
        and region.connection_upper_y is not None
        and region_outer.connection_upper_y is None
    ):
        # da_outer may have more points in the y-direction, because it has an actual
        # boundary, not a connection. Need to remove any extra points
        da_outer = da_outer.isel(**{ycoord: slice(None, -myg_da)})

    # select just the points we need to fill the guard cells of da
    da_outer = da_outer.isel(**{xcoord: slice(mxg)})

    if (
        myg_da > 0
        and keep_yboundaries
        and region.connection_lower_y is None
        and region_outer.connection_lower_y is not None
    ):
        # da_outer may have fewer points in the y-direction, because it has a connection,
        # not an actual boundary. Need to get the extra points from its connection
        da_outer_lower = da_global.bout.from_region(
            region_outer.connection_lower_y, with_guards=0
        )
        da_outer_lower = da_outer_lower.isel(
            **{xcoord: slice(-mxg, None), ycoord: slice(-myg_da, None)}
        )
        save_regions = da_outer.bout._regions
        da_outer = xr.concat((da_outer_lower, da_outer), ycoord, join="exact")
        # xr.concat takes attributes from the first variable, but we need da_outer's
        # regions
        da_outer.attrs["regions"] = save_regions
    if (
        myg_da > 0
        and keep_yboundaries
        and region.connection_upper_y is None
        and region_outer.connection_upper_y is not None
    ):
        # da_outer may have fewer points in the y-direction, because it has a connection,
        # not an actual boundary. Need to get the extra points from its connection
        da_outer_upper = da_global.bout.from_region(
            region_outer.connection_upper_y, with_guards=0
        )
        da_outer_upper = da_outer_upper.isel(
            **{xcoord: slice(-mxg, None), ycoord: slice(myg_da)}
        )
        da_outer = xr.concat((da_outer, da_outer_upper), ycoord, join="exact")
        # xr.concat takes attributes from the first variable, so regions are OK

    if xcoord in da.coords:
        # Some coordinates may not be single-valued, so use local coordinates for
        # neighbouring region, not communicated ones. Ensures the coordinates are
        # continuous so that interpolation works correctly near the boundaries.
        slices = region.get_outer_guards_slices(mxg=mxg)
        new_xcoord = da_global[xcoord].isel(**{xcoord: slices[xcoord]})
        new_ycoord = da_global[ycoord].isel(**{ycoord: slices[ycoord]})

        # can't use commented out version, uncommented one works around xarray bug
        # removing attrs
        # https://github.com/pydata/xarray/issues/4415
        # https://github.com/pydata/xarray/issues/4393
        # da_outer = da_outer.assign_coords(**{xcoord: new_xcoord, ycoord: new_ycoord})
        da_outer[xcoord].data[...] = new_xcoord.data
        da_outer[ycoord].data[...] = new_ycoord.data

    save_regions = da.bout._regions
    da = xr.concat((da, da_outer), xcoord, join="exact")
    # xarray<0.15.0 only keeps attrs that are the same on all variables passed to concat
    da.attrs["regions"] = save_regions

    return da


def _concat_lower_guards(da, da_global, mxg, myg):
    """
    Concatenate lower y-guard cells to da, which is in a single region, getting the guard
    cells from the global array
    """

    if myg <= 0:
        return da

    if len(da.bout._regions) > 1:
        raise ValueError("da passed should have only one region")
    region = list(da.bout._regions.values())[0]

    if region.connection_lower_y not in da_global.bout._regions:
        # No connection, or plotting restricted set of regions not including this
        # connection
        return da

    xcoord = da_global.metadata["bout_xdim"]
    ycoord = da_global.metadata["bout_ydim"]

    da_lower = da_global.bout.from_region(
        region.connection_lower_y, with_guards={xcoord: mxg, ycoord: 0}
    )

    # select just the points we need to fill the guard cells of da
    da_lower = da_lower.isel(**{ycoord: slice(-myg, None)})

    if ycoord in da.coords:
        # Some coordinates may not be single-valued, so use local coordinates for
        # neighbouring region, not communicated ones. Ensures the coordinates are
        # continuous so that interpolation works correctly near the boundaries.
        slices = region.get_lower_guards_slices(mxg=mxg, myg=myg)

        new_xcoord = da_global[xcoord].isel(**{xcoord: slices[xcoord]})
        new_ycoord = da_global[ycoord].isel(**{ycoord: slices[ycoord]})

        # new_ycoord might not join smoothly onto da[ycoord] for limiter or core-only
        # geometries where the guard cells have to come from the opposite side of a
        # branch cut. Need to work around this, or it breaks parallel interpolation.
        ycoord_increasing = np.any(
            da[ycoord].isel(**{ycoord: 1}) > da[ycoord].isel(**{ycoord: 0})
        )

        if ycoord_increasing and np.any(
            new_ycoord.isel(**{ycoord: -1}) > da[ycoord].isel(**{ycoord: 0})
        ):
            if not np.all(
                new_ycoord.isel(**{ycoord: -1}) > da[ycoord].isel(**{ycoord: 0})
            ):
                raise ValueError(
                    f"Inconsistent {ycoord} - should be either increasing everywhere "
                    f"or decreasing everywhere"
                )
            # Estimate what coordinate values should be by extrapolating linearly from
            # interior of region. This is exact if the grid spacing is constant, which
            # it usually is in 'y'
            offset = (
                2.0 * da[ycoord].isel(**{ycoord: 0})
                - da[ycoord].isel(**{ycoord: 1})
                - new_ycoord.isel(**{ycoord: -1})
            )
            new_ycoord = new_ycoord + offset

        elif (not ycoord_increasing) and np.any(
            new_ycoord.isel(**{ycoord: -1}) < da[ycoord].isel(**{ycoord: 0})
        ):
            if not np.all(
                new_ycoord.isel(**{ycoord: -1}) < da[ycoord].isel(**{ycoord: 0})
            ):
                raise ValueError(
                    f"Inconsistent {ycoord} - should be either increasing everywhere "
                    f"or decreasing everywhere"
                )
            # Estimate what coordinate values should be by extrapolating linearly from
            # interior of region. This is exact if the grid spacing is constant, which
            # it usually is in 'y'
            offset = (
                2.0 * da[ycoord].isel(**{ycoord: 0})
                - da[ycoord].isel(**{ycoord: 1})
                - new_ycoord.isel(**{ycoord: -1})
            )
            new_ycoord = new_ycoord + offset

        # can't use commented out version, uncommented one works around xarray bug
        # removing attrs
        # https://github.com/pydata/xarray/issues/4415
        # https://github.com/pydata/xarray/issues/4393
        # da_lower = da_lower.assign_coords(**{xcoord: new_xcoord, ycoord: new_ycoord})
        da_lower[xcoord].data[...] = new_xcoord.data
        da_lower[ycoord].data[...] = new_ycoord.data

    if "poloidal_distance" in da.coords and myg > 0:
        # Special handling for core regions to deal with branch cut
        if "core" in region.name:
            # import pdb; pdb.set_trace()
            # Try to detect whether there is branch cut at lower boundary: if there is
            # poloidal_distance_ylow should be zero at the boundary of this region
            poloidal_distance_bottom = da["poloidal_distance_ylow"].isel({ycoord: 0})
            if all(abs(poloidal_distance_bottom) < 1.0e-16):
                # Offset so that the poloidal_distance in da_lower is continuous from
                # the poloidal_distance in this region.
                # Expect there to be y-boundary cells in the Dataset, this will probably
                # fail if there are not.
                total_poloidal_distance = da["total_poloidal_distance"]
                da_lower["poloidal_distance"] -= total_poloidal_distance
                da_lower["poloidal_distance_ylow"] -= total_poloidal_distance

    save_regions = da.bout._regions
    da = xr.concat((da_lower, da), ycoord, join="exact")
    # xr.concat takes attributes from the first variable (for xarray>=0.15.0, keeps attrs
    # that are the same in all objects for xarray<0.15.0)
    da.attrs["regions"] = save_regions

    return da


def _concat_upper_guards(da, da_global, mxg, myg):
    """
    Concatenate upper y-guard cells to da, which is in a single region, getting the guard
    cells from the global array
    """

    if myg <= 0:
        return da

    if len(da.bout._regions) > 1:
        raise ValueError("da passed should have only one region")
    region = list(da.bout._regions.values())[0]

    if region.connection_upper_y not in da_global.bout._regions:
        # No connection, or plotting restricted set of regions not including this
        # connection
        return da

    xcoord = da_global.metadata["bout_xdim"]
    ycoord = da_global.metadata["bout_ydim"]

    da_upper = da_global.bout.from_region(
        region.connection_upper_y, with_guards={xcoord: mxg, ycoord: 0}
    )
    # select just the points we need to fill the guard cells of da
    da_upper = da_upper.isel(**{ycoord: slice(myg)})

    if ycoord in da.coords:
        # Some coordinates may not be single-valued, so use local coordinates for
        # neighbouring region, not communicated ones. Ensures the coordinates are
        # continuous so that interpolation works correctly near the boundaries.
        slices = region.get_upper_guards_slices(mxg=mxg, myg=myg)

        new_xcoord = da_global[xcoord].isel(**{xcoord: slices[xcoord]})
        new_ycoord = da_global[ycoord].isel(**{ycoord: slices[ycoord]})

        # new_ycoord might not join smoothly onto da[ycoord] for limiter or core-only
        # geometries where the guard cells have to come from the opposite side of a
        # branch cut. Need to work around this, or it breaks parallel interpolation.
        ycoord_increasing = np.any(
            da[ycoord].isel(**{ycoord: -1}) > da[ycoord].isel(**{ycoord: -2})
        )

        if ycoord_increasing and np.any(
            new_ycoord.isel(**{ycoord: 0}) < da[ycoord].isel(**{ycoord: -1})
        ):
            if not np.all(
                new_ycoord.isel(**{ycoord: 0}) < da[ycoord].isel(**{ycoord: -1})
            ):
                raise ValueError(
                    f"Inconsistent {ycoord} - should be either increasing everywhere "
                    f"or decreasing everywhere"
                )
            # Estimate what coordinate values should be by extrapolating linearly from
            # interior of region. This is exact if the grid spacing is constant, which
            # it usually is in 'y'
            offset = (
                2.0 * da[ycoord].isel(**{ycoord: -1})
                - da[ycoord].isel(**{ycoord: -2})
                - new_ycoord.isel(**{ycoord: 0})
            )
            new_ycoord = new_ycoord + offset

        elif (not ycoord_increasing) and np.any(
            new_ycoord.isel(**{ycoord: 0}) > da[ycoord].isel(**{ycoord: -1})
        ):
            if not np.all(
                new_ycoord.isel(**{ycoord: 0}) > da[ycoord].isel(**{ycoord: -1})
            ):
                raise ValueError(
                    f"Inconsistent {ycoord} - should be either increasing everywhere "
                    f"or decreasing everywhere"
                )
            # Estimate what coordinate values should be by extrapolating linearly from
            # interior of region. This is exact if the grid spacing is constant, which
            # it usually is in 'y'
            offset = (
                2.0 * da[ycoord].isel(**{ycoord: -1})
                - da[ycoord].isel(**{ycoord: -2})
                - new_ycoord.isel(**{ycoord: 0})
            )
            new_ycoord = new_ycoord + offset

        # can't use commented out version, uncommented one works around xarray bug
        # removing attrs
        # https://github.com/pydata/xarray/issues/4415
        # https://github.com/pydata/xarray/issues/4393
        # da_upper = da_upper.assign_coords(**{xcoord: new_xcoord, ycoord: new_ycoord})
        da_upper[xcoord].data[...] = new_xcoord.data
        da_upper[ycoord].data[...] = new_ycoord.data

    if "poloidal_distance" in da.coords and myg > 0:
        # Special handling for core regions to deal with branch cut
        if "core" in region.name:
            # import pdb; pdb.set_trace()
            # Try to detect whether there is branch cut at upper boundary: if there is
            # poloidal_distance_ylow should be zero at the boundary of da_upper
            poloidal_distance_bottom = da_upper["poloidal_distance_ylow"].isel(
                {ycoord: 0}
            )
            if all(abs(poloidal_distance_bottom) < 1.0e-16):
                # Offset so that the poloidal_distance in da_upper is continuous from
                # the poloidal_distance in this region.
                # Expect there to be y-boundary cells in the Dataset, this will probably
                # fail if there are not.
                total_poloidal_distance = da["total_poloidal_distance"]
                da_upper["poloidal_distance"] += total_poloidal_distance
                da_upper["poloidal_distance_ylow"] += total_poloidal_distance

    save_regions = da.bout._regions
    da = xr.concat((da, da_upper), ycoord, join="exact")
    # xarray<0.15.0 only keeps attrs that are the same on all variables passed to concat
    da.attrs["regions"] = save_regions

    return da


def _from_region(ds_or_da, name, with_guards):
    # ensure we do not modify the input
    ds_or_da = ds_or_da.copy(deep=True)

    region = ds_or_da.bout._regions[name]
    xcoord = ds_or_da.metadata["bout_xdim"]
    ycoord = ds_or_da.metadata["bout_ydim"]

    if with_guards is None:
        mxg = ds_or_da.metadata["MXG"]
        myg = ds_or_da.metadata["MYG"]
    else:
        try:
            mxg = with_guards.get(xcoord, ds_or_da.metadata["MXG"])
            myg = with_guards.get(ycoord, ds_or_da.metadata["MYG"])
        except AttributeError:
            mxg = with_guards
            myg = with_guards

    result = ds_or_da.isel(region.get_slices()).copy()

    # The returned result has only one region
    single_region = deepcopy(region)
    result.attrs["regions"] = {name: single_region}

    # get inner x-guard cells for result from the global array
    result = _concat_inner_guards(result, ds_or_da, mxg)
    # get outer x-guard cells for result from the global array
    result = _concat_outer_guards(result, ds_or_da, mxg)
    # get lower y-guard cells from the global array
    result = _concat_lower_guards(result, ds_or_da, mxg, myg)
    # get upper y-guard cells from the global array
    result = _concat_upper_guards(result, ds_or_da, mxg, myg)

    # If the result (which only has a single region) is passed to from_region a
    # second time, don't want to slice anything.
    single_region = list(result.bout._regions.values())[0]
    single_region.xinner_ind = None
    single_region.xouter_ind = None
    single_region.ylower_ind = None
    single_region.yupper_ind = None

    return result



from collections.abc import Sequence
import matplotlib
import matplotlib.pyplot as plt
import numpy as np
from pathlib import Path
from tempfile import TemporaryDirectory
import warnings

import xarray as xr

from .utils import (
    _create_norm,
    _decompose_regions,
    _is_core_only,
    _k3d_plot_isel,
    plot_separatrices,
    plot_targets,
)


if (
    "pcolor.shading" in matplotlib.rcParams
    and matplotlib.rcParams["pcolor.shading"] == "flat"
):
    # "flat" was the old matplotlib default which discarded the last row and column if
    # X, Y and Z were all equal in size. The new "auto" should be better. Need to set
    # this explicitly because "flat" is the default during a deprecation cycle.
    matplotlib.rcParams["pcolor.shading"] = "auto"


def plot_regions(da, ax=None, **kwargs):
    """
    Plots each logical plotting region as a different color for debugging.

    Uses matplotlib.pcolormesh
    """

    x = kwargs.pop("x", "R")
    y = kwargs.pop("y", "Z")

    if len(da.dims) != 2:
        raise ValueError("da must be 2D (x,y)")

    if ax is None:
        fig, ax = plt.subplots()

    da_regions = _decompose_regions(da)

    colored_regions = [
        xr.full_like(da_region, fill_value=num / len(da_regions))
        for num, da_region in enumerate(da_regions.values())
    ]

    with warnings.catch_warnings():
        # The coordinates we pass are a logically rectangular grid, so should be fine
        # even if this warning is triggered.
        warnings.filterwarnings(
            "ignore",
            "The input coordinates to pcolormesh are interpreted as cell centers, but "
            "are not monotonically increasing or decreasing. This may lead to "
            "incorrectly calculated cell edges, in which case, please supply explicit "
            "cell edges to pcolormesh.",
            UserWarning,
        )
        result = [
            region.plot.pcolormesh(
                x=x,
                y=y,
                vmin=0,
                vmax=1,
                cmap="tab20",
                infer_intervals=False,
                add_colorbar=False,
                ax=ax,
                **kwargs,
            )
            for region in colored_regions
        ]

    return result


def plot2d_wrapper(
    da,
    method,
    *,
    ax=None,
    separatrix=True,
    targets=True,
    add_limiter_hatching=True,
    gridlines=None,
    cmap=None,
    norm=None,
    logscale=None,
    vmin=None,
    vmax=None,
    aspect=None,
    extend=None,
    **kwargs,
):
    """
    Make a 2D plot using an xarray method, taking into account branch cuts (X-points).

    Wraps `xarray.DataArray.plot` methods, so automatically adds labels.

    Parameters
    ----------
    da : xarray.DataArray
        A 2D (x,y) DataArray of data to plot
    method : xarray.plot.*
        An xarray plotting method to use
    ax : Axes, optional
        A matplotlib axes instance to plot to. If None, create a new
        figure and axes, and plot to that
    separatrix : bool, optional
        Add dashed lines showing separatrices
    targets : bool, optional
        Draw solid lines at the target surfaces
    add_limiter_hatching : bool, optional
        Draw hatched areas at the targets
    gridlines : bool, int or slice or dict of bool, int or slice, optional
        If True, draw grid lines on the plot. If an int is passed, it is used as the
        stride when plotting grid lines (to reduce the number on the plot). If a slice is
        passed it is used to select the grid lines to plot.
        If a dict is passed, the 'x' entry (bool, int or slice) is used for the radial
        grid-lines and the 'y' entry for the poloidal grid lines.
    cmap : Matplotlib colormap, optional
        Color map to use for the plot
    norm : matplotlib.colors.Normalize instance, optional
        Normalization to use for the color scale.
        Cannot be set at the same time as 'logscale'
    logscale : bool or float, optional
        If True, default to a logarithmic color scale instead of a linear one.
        If a non-bool type is passed it is treated as a float used to set the linear
        threshold of a symmetric logarithmic scale as
        linthresh=min(abs(vmin),abs(vmax))*logscale, defaults to 1e-5 if True is passed.
        Cannot be set at the same time as 'norm'
    vmin : float, optional
        Minimum value for the color scale
    vmax : float, optional
        Maximum value for the color scale
    aspect : str or float, optional
        Passed to ax.set_aspect(). By default 'equal' is used.
    extend : str or None, optional
        Passed to fig.colorbar()
    levels : int or iterable, optional
        Only used by contour or contourf, sets the number of levels (if int) or the level
        values (if iterable)
    **kwargs : optional
        Additional arguments are passed on to method

    Returns
    -------
    artists
        List of the artist instances
    """

    # TODO generalise this
    x = kwargs.pop("x", "R")
    y = kwargs.pop("y", "Z")

    if len(da.dims) != 2:
        raise ValueError("da must be 2D (x,y)")

    if ax is None:
        fig, ax = plt.subplots()
    else:
        fig = ax.get_figure()

    if aspect is None:
        aspect = "equal"
    ax.set_aspect(aspect)

    if vmin is None:
        vmin = da.min().values
    if vmax is None:
        vmax = da.max().values

    if extend is None:
        # Replicate default for older matplotlib that does not handle extend=None
        # matplotlib-3.3 definitely does not need this. Not sure about 3.0, 3.1, 3.2.
        extend = "neither"

    # set up 'levels' if needed
    if method is xr.plot.contourf or method is xr.plot.contour:
        levels = kwargs.get("levels", 7)
        if isinstance(levels, Sequence):
            levels = np.array(list(levels))
            kwargs["levels"] = levels
            vmin = np.min(levels)
            vmax = np.max(levels)
        else:
            levels = np.linspace(vmin, vmax, levels, endpoint=True)
            # put levels back into kwargs
            kwargs["levels"] = levels

        levels = kwargs.get("levels", 7)
        if isinstance(levels, np.int64):
            levels = np.linspace(vmin, vmax, levels, endpoint=True)
            # put levels back into kwargs
            kwargs["levels"] = levels
        else:
            levels = np.array(list(levels))
            kwargs["levels"] = levels
            vmin = np.min(levels)
            vmax = np.max(levels)

    # Need to create a colorscale that covers the range of values in the whole array.
    # Using the add_colorbar argument would create a separate color scale for each
    # separate region, which would not make sense.
    if method is xr.plot.contourf:
        # create colorbar
        norm = _create_norm(logscale, norm, vmin, vmax)
        sm = plt.cm.ScalarMappable(norm=norm, cmap=cmap)
        # make colorbar have only discrete levels
        # average the levels so that colors in the colorbar represent the intervals
        # between the levels, as contourf colors filled regions between the given levels.
        cmap = matplotlib.colors.LinearSegmentedColormap.from_list(
            "discrete cmap",
            sm.to_rgba(0.5 * (levels[:-1] + levels[1:])),
            len(levels) - 1,
        )
        # re-make sm with new cmap
        sm = plt.cm.ScalarMappable(norm=norm, cmap=cmap)
        sm.set_array([])
        fig.colorbar(sm, ticks=levels, ax=ax, extend=extend)
    elif method is xr.plot.contour:
        # create colormap to be shared by all regions
        norm = matplotlib.colors.Normalize(vmin=levels[0], vmax=levels[-1])
        sm = plt.cm.ScalarMappable(norm=norm, cmap=cmap)
        cmap = matplotlib.colors.ListedColormap(
            sm.to_rgba(levels), name="discrete cmap"
        )
    else:
        # pass vmin and vmax through kwargs as they are not used for contourf or contour
        # plots
        kwargs["vmin"] = vmin
        kwargs["vmax"] = vmax

        # create colorbar
        norm = _create_norm(logscale, norm, vmin, vmax)
        sm = plt.cm.ScalarMappable(norm=norm, cmap=cmap)
        sm.set_array([])
        cmap = sm.get_cmap()
        cbar = fig.colorbar(sm, ax=ax, extend=extend)
        if "long_name" in da.attrs:
            cbar_label = da.long_name
        else:
            cbar_label = da.name
        if "units" in da.attrs:
            cbar_label += f" [{da.units}]"
        cbar.ax.set_ylabel(cbar_label)

    if method is xr.plot.pcolormesh:
        if "infer_intervals" not in kwargs:
            kwargs["infer_intervals"] = False

    da_regions = _decompose_regions(da)

    # Plot all regions on same axis
    add_labels = [True] + [False] * (len(da_regions) - 1)
    with warnings.catch_warnings():
        # The coordinates we pass are a logically rectangular grid, so should be fine
        # even if this warning is triggered by pcolor or pcolormesh
        warnings.filterwarnings(
            "ignore",
            "The input coordinates to pcolormesh are interpreted as cell centers, but "
            "are not monotonically increasing or decreasing. This may lead to "
            "incorrectly calculated cell edges, in which case, please supply explicit "
            "cell edges to pcolormesh.",
            UserWarning,
        )
        artists = [
            method(
                region,
                x=x,
                y=y,
                ax=ax,
                add_colorbar=False,
                add_labels=add_label,
                cmap=cmap,
                norm=norm,
                **kwargs,
            )
            for region, add_label in zip(da_regions.values(), add_labels)
        ]

    if method is xr.plot.contour:
        fig.colorbar(artists[0], ax=ax)

    if gridlines is not None:
        # convert gridlines to dict
        if not isinstance(gridlines, dict):
            gridlines = {"x": gridlines, "y": gridlines}

        for key, value in gridlines.items():
            if value is True:
                gridlines[key] = slice(None)
            elif isinstance(value, int):
                gridlines[key] = slice(0, None, value)
            elif value is not None:
                if not isinstance(value, slice):
                    raise ValueError(
                        "Argument passed to gridlines must be bool, int or "
                        "slice. Got a " + type(value) + ", " + str(value)
                    )

        x_regions = [da_region[x] for da_region in da_regions.values()]
        y_regions = [da_region[y] for da_region in da_regions.values()]

        for x, y in zip(x_regions, y_regions):
            if (
                da.metadata["bout_xdim"] not in x.dims
                and da.metadata["bout_ydim"] not in x.dims
            ) or (
                da.metadata["bout_xdim"] not in y.dims
                and da.metadata["bout_ydim"] not in y.dims
            ):
                # Small regions around X-point do not have segments in x- or y-directions,
                # so skip
                # Currently this region does not exist, but there is a small white gap at
                # the X-point, so we might add it back in future
                continue
            if gridlines.get("x") is not None:
                # transpose in case Dataset or DataArray has been transposed away from the usual
                # form
                dim_order = (da.metadata["bout_xdim"], da.metadata["bout_ydim"])
                yarg = {da.metadata["bout_ydim"]: gridlines["x"]}
                plt.plot(
                    x.isel(**yarg).transpose(*dim_order, transpose_coords=True),
                    y.isel(**yarg).transpose(*dim_order, transpose_coords=True),
                    color="k",
                    lw=0.1,
                )
            if gridlines.get("y") is not None:
                xarg = {da.metadata["bout_xdim"]: gridlines["y"]}
                # Need to plot transposed arrays to make gridlines that go in the
                # y-direction
                dim_order = (da.metadata["bout_ydim"], da.metadata["bout_xdim"])
                plt.plot(
                    x.isel(**xarg).transpose(*dim_order, transpose_coords=True),
                    y.isel(**yarg).transpose(*dim_order, transpose_coords=True),
                    color="k",
                    lw=0.1,
                )

    ax.set_title(da.name)

    if _is_core_only(da):
        separatrix = False
        targets = False

    if separatrix:
        plot_separatrices(da_regions, ax, x=x, y=y)

    if targets:
        plot_targets(da_regions, ax, x=x, y=y, hatching=add_limiter_hatching)

    return artists


def plot3d(
    da,
    style="surface",
    engine="k3d",
    levels=None,
    outputgrid=(100, 100, 25),
    color_map=None,
    colorbar=True,
    colorbar_font_size=None,
    plot=None,
    save_as=None,
    surface_xinds=None,
    surface_yinds=None,
    surface_zinds=None,
    fps=20.0,
    mayavi_figure=None,
    mayavi_figure_args=None,
    mayavi_view=None,
    **kwargs,
):
    """
    Make a 3d plot

    Warnings
    --------

    3d plotting functionality is still a bit of a work in progress. Bugs are likely, and
    help developing is welcome!

    Parameters
    ----------
    style : {'surface', 'poloidal planes'}
        Type of plot to make:
        - 'surface' plots the outer surface of the DataArray
        - 'poloidal planes' plots each poloidal plane in the DataArray
    engine : {'k3d', 'mayavi'}
        3d plotting library to use
    levels : sequence of (float, float)
        For isosurface, the pairs of (level-value, opacity) to plot
    outputgrid : (int, int, int) or None, optional
        For isosurface or volume plots, the number of points to use in the Cartesian
        (X,Y,Z) grid, that data is interpolated onto for plotting. If None, then do not
        interpolate and treat "bout_xdim", "bout_ydim" and "bout_zdim" coordinates as
        Cartesian (useful for slab simulations).
    color_map : k3d color map, optional
        Color map for k3d plots
    colorbar : bool or dict, default True
        Add a color bar. If a dict is passed, it is passed on to the colorbar() function
        as keyword arguments.
    colorbar_font_size : float, default None
        Set the font size used by the colorbar (for engine="mayavi")
    plot : k3d plot instance, optional
        Existing plot to add new plots to
    save_as : str
        Filename to save figure to. Animations will be saved as a sequence of
        numbered files.
    surface_xinds : (int, int), default None
        Indices to select when plotting radial surfaces. These indices are local to the
        region being plotted, so values will be strange. Recommend using values relative
        to the radial boundaries (i.e. positive for inner boundary and negative for
        outer boundary).
    surface_yinds : (int, int), default None
        Indices to select when plotting poloidal surfaces. These indices are local to the
        region being plotted, so values will be strange. Recommend using values relative
        to the poloidal boundaries (i.e. positive for lower boundaries and negative for
        upper boundaries).
    surface_zinds : (int, int), default None
        Indices to select when plotting toroidal surfaces
    fps : float, default 20
        Frames per second to use when creating an animation.
    mayavi_figure : mayavi.core.scene.Scene, default None
        Existing Mayavi figure to add this plot to.
    mayavi_figure_args : dict, default None
        Arguments to use when creating a new Mayavi figure. Ignored if ``mayavi_figure``
        is passed.
    mayavi_view : (float, float, float), default None
        If set, arguments are passed to `mayavi.mlab.view` to set the
        view when ``engine="mayavi"``
    vmin, vmax : float
        vmin and vmax are treated specially. If a float is passed, then it is used for
        vmin/vmax. If the arguments are not passed, then the minimum and maximum of the
        data are used. For an animation, to get minimum and/or maximum calculated
        separately for each frame, pass ``vmin=None`` and/or ``vmax=None`` explicitly.
    **kwargs
        Extra keyword arguments are passed to the backend plotting function
    """

    tcoord = da.metadata["bout_tdim"]
    xcoord = da.metadata["bout_xdim"]
    ycoord = da.metadata["bout_ydim"]
    zcoord = da.metadata["bout_zdim"]

    if tcoord in da.dims:
        animate = True
        if len(da.dims) != 4:
            raise ValueError(
                f"plot3d needs to be passed 3d spatial data. Got {da.dims}."
            )
    else:
        animate = False
        if len(da.dims) != 3:
            raise ValueError(
                f"plot3d needs to be passed 3d spatial data. Got {da.dims}."
            )

    da = da.bout.add_cartesian_coordinates()
    if "vmin" in kwargs:
        vmin = kwargs.pop("vmin")
    else:
        vmin = float(da.min().values)
    if "vmax" in kwargs:
        vmax = kwargs.pop("vmax")
    else:
        vmax = float(da.max().values)

    if engine == "k3d":
        if animate:
            raise ValueError(
                "animation not supported by k3d, do not pass time-dependent DataArray"
            )
        if save_as is not None:
            raise ValueError("save_as not supported by k3d implementation yet")
        if colorbar:
            warnings.warn("colorbar not added to k3d plots yet")

        try:
            import k3d
        except ImportError:
            raise ImportError(
                'Please install the `k3d` package for 3d plotting with `engine="k3d"`'
            )

        if color_map is None:
            color_map = k3d.matplotlib_color_maps.Viridis

        if plot is None:
            plot = k3d.plot()
            return_plot = True
        else:
            return_plot = False

        if style == "isosurface" or style == "volume":
            data = da.copy(deep=True).load()
            datamin = data.min().item()
            datamax = data.max().item()

            if outputgrid is None:
                Xmin = da[da.metadata["bout_xdim"]][0]
                Xmax = da[da.metadata["bout_xdim"]][-1]
                Ymin = da[da.metadata["bout_ydim"]][0]
                Ymax = da[da.metadata["bout_ydim"]][-1]
                Zmin = da[da.metadata["bout_zdim"]][0]
                Zmax = da[da.metadata["bout_zdim"]][-1]

                grid = da.astype(np.float32).values
            else:
                xpoints, ypoints, zpoints = outputgrid
                nx, ny, nz = data.shape

                # interpolate to Cartesian array
                Xmin = data["X_cartesian"].min()
                Xmax = data["X_cartesian"].max()
                Ymin = data["Y_cartesian"].min()
                Ymax = data["Y_cartesian"].max()
                Zmin = data["Z_cartesian"].min()
                Zmax = data["Z_cartesian"].max()
                Rmin = data["R"].min()
                Rmax = data["R"].max()
                Zmin = data["Z"].min()
                Zmax = data["Z"].max()
                newX = xr.DataArray(
                    np.linspace(Xmin, Xmax, xpoints), dims="x"
                ).expand_dims({"y": ypoints, "z": zpoints}, axis=[1, 0])
                newY = xr.DataArray(
                    np.linspace(Ymin, Ymax, ypoints), dims="y"
                ).expand_dims({"x": xpoints, "z": zpoints}, axis=[2, 0])
                newZ = xr.DataArray(
                    np.linspace(Zmin, Zmax, zpoints), dims="z"
                ).expand_dims({"x": xpoints, "y": ypoints}, axis=[2, 1])
                newR = np.sqrt(newX**2 + newY**2)
                newzeta = np.arctan2(newY, newX)  # .values

                from scipy.interpolate import (
                    RegularGridInterpolator,
                    griddata,
                )

                print("start interpolating")
                Rcyl = xr.DataArray(
                    np.linspace(Rmin, Rmax, 2 * zpoints), dims="r"
                ).expand_dims({"z": 2 * zpoints}, axis=1)
                Zcyl = xr.DataArray(
                    np.linspace(Zmin, Zmax, 2 * zpoints), dims="z"
                ).expand_dims({"r": 2 * zpoints}, axis=0)

                # Interpolate in two stages for efficiency. Unstructured 3d interpolation is
                # very slow. Unstructured 2d interpolation onto Cartesian (R, Z) grids,
                # followed by structured 3d interpolation onto the (X, Y, Z) grid, is much
                # faster.
                # Structured 3d interpolation straight from (psi, theta, zeta) to (X, Y, Z)
                # leaves artifacts in the output, because theta does not vary continuously
                # everywhere (has branch cuts).

                # order of dimensions does not really matter here - output only depends on
                # shape of newR, newZ, newzeta. Possibly more efficient to assign the 2d
                # results in the loop to the last two dimensions, so put zeta first.
                data_cyl = np.zeros((nz, 2 * zpoints, 2 * zpoints))
                print("interpolate poloidal planes")
                for z in range(nz):
                    data_cyl[z] = griddata(
                        (data["R"].values.flatten(), data["Z"].values.flatten()),
                        data.isel(zeta=z).values.flatten(),
                        (Rcyl.values, Zcyl.values),
                        method="cubic",
                        fill_value=datamin - 2.0 * (datamax - datamin),
                    )
                print("build 3d interpolator")
                interp = RegularGridInterpolator(
                    (data["zeta"].values, Rcyl.isel(z=0).values, Zcyl.isel(r=0).values),
                    data_cyl,
                    bounds_error=False,
                    fill_value=datamin - 2.0 * (datamax - datamin),
                )
                print("do 3d interpolation")
                grid = interp(
                    (newzeta, newR, newZ),
                    method="linear",
                )
                print("done interpolating")

            if style == "isosurface":
                if levels is None:
                    levels = [(0.5 * (datamin + datamax), 1.0)]
                for level, opacity in levels:
                    plot += k3d.marching_cubes(
                        grid.astype(np.float32),
                        bounds=[Xmin, Xmax, Ymin, Ymax, Zmin, Zmax],
                        level=level,
                        color_map=color_map,
                    )
            elif style == "volume":
                plot += k3d.volume(
                    grid.astype(np.float32),
                    color_range=[datamin, datamax],
                    bounds=[Xmin, Xmax, Ymin, Ymax, Zmin, Zmax],
                    color_map=color_map,
                )
            if return_plot:
                return plot
            else:
                return

        for region_name, da_region in _decompose_regions(da).items():
            npsi, ntheta, nzeta = da_region.shape

            if style == "surface":
                region = da_region.regions[region_name]

                if region.connection_inner_x is None:
                    # Plot the inner-x surface
                    plot += _k3d_plot_isel(
                        da_region,
                        {xcoord: 0},
                        vmin,
                        vmax,
                        color_map=color_map,
                        **kwargs,
                    )

                if region.connection_outer_x is None:
                    # Plot the outer-x surface
                    plot += _k3d_plot_isel(
                        da_region,
                        {xcoord: -1},
                        vmin,
                        vmax,
                        color_map=color_map,
                        **kwargs,
                    )

                if region.connection_lower_y is None:
                    # Plot the lower-y surface
                    plot += _k3d_plot_isel(
                        da_region,
                        {ycoord: 0},
                        vmin,
                        vmax,
                        color_map=color_map,
                        **kwargs,
                    )

                if region.connection_upper_y is None:
                    # Plot the upper-y surface
                    plot += _k3d_plot_isel(
                        da_region,
                        {ycoord: -1},
                        vmin,
                        vmax,
                        color_map=color_map,
                        **kwargs,
                    )

                # First z-surface
                plot += _k3d_plot_isel(
                    da_region, {zcoord: 0}, vmin, vmax, color_map=color_map, **kwargs
                )

                # Last z-surface
                plot += _k3d_plot_isel(
                    da_region, {zcoord: -1}, vmin, vmax, color_map=color_map, **kwargs
                )
            elif style == "poloidal planes":
                for zeta in range(nzeta):
                    plot += _k3d_plot_isel(
                        da_region,
                        {zcoord: zeta},
                        vmin,
                        vmax,
                        color_map=color_map,
                        **kwargs,
                    )
            else:
                raise ValueError(f"style='{style}' not implemented for engine='k3d'")

        if return_plot:
            return plot
        else:
            return

    elif engine == "mayavi":
        try:
            from mayavi import mlab
        except ImportError:
            raise ImportError(
                "Please install the `mayavi` package for 3d plotting with "
                '`engine="mayavi"`'
            )

        if mayavi_figure is None:
            if mayavi_figure_args is not None:
                mlab.figure(**mayavi_figure_args)
        else:
            mlab.figure(mayavi_figure)

        if style == "surface":

            def create_or_update_plot(plot_objects=None, tind=None, this_save_as=None):
                if plot_objects is None:
                    # Creating plot for first time
                    plot_objects_to_return = {}
                    this_da = da
                if tind is not None:
                    this_da = da.isel({tcoord: tind})

                for region_name, da_region in _decompose_regions(this_da).items():
                    region = da_region.regions[region_name]

                    # Always include z-surfaces
                    zstart_ind = 0 if surface_zinds is None else surface_zinds[0]
                    zend_ind = -1 if surface_zinds is None else surface_zinds[1]
                    surface_selections = [
                        {this_da.metadata["bout_zdim"]: zstart_ind},
                        {this_da.metadata["bout_zdim"]: zend_ind},
                    ]
                    if region.connection_inner_x is None:
                        # Plot the inner-x surface
                        xstart_ind = 0 if surface_xinds is None else surface_xinds[0]
                        surface_selections.append({xcoord: xstart_ind})
                    if region.connection_outer_x is None:
                        # Plot the outer-x surface
                        xend_ind = -1 if surface_xinds is None else surface_xinds[1]
                        surface_selections.append({xcoord: xend_ind})
                    if region.connection_lower_y is None:
                        # Plot the lower-y surface
                        ystart_ind = 0 if surface_yinds is None else surface_yinds[0]
                        surface_selections.append({ycoord: ystart_ind})
                    if region.connection_upper_y is None:
                        # Plot the upper-y surface
                        yend_ind = -1 if surface_yinds is None else surface_yinds[1]
                        surface_selections.append({ycoord: yend_ind})

                    for i, surface_sel in enumerate(surface_selections):
                        da_sel = da_region.isel(surface_sel)
                        X = da_sel["X_cartesian"].values
                        Y = da_sel["Y_cartesian"].values
                        Z = da_sel["Z_cartesian"].values
                        data = da_sel.values

                        if plot_objects is None:
                            plot_objects_to_return[region_name + str(i)] = mlab.mesh(
                                X, Y, Z, scalars=data, vmin=vmin, vmax=vmax, **kwargs
                            )
                        else:
                            plot_objects[
                                region_name + str(i)
                            ].mlab_source.scalars = data

                if mayavi_view is not None:
                    mlab.view(*mayavi_view)

                if colorbar and (tind is None or tind == 0):
                    if isinstance(colorbar, dict):
                        colorbar_args = colorbar
                    else:
                        colorbar_args = {}
                    cb = mlab.colorbar(**colorbar_args)
                    if colorbar_font_size is not None:
                        cb.scalar_bar.unconstrained_font_size = True
                        cb.label_text_property.font_size = colorbar_font_size
                        cb.title_text_property.font_size = colorbar_font_size

                if this_save_as:
                    if tind is None:
                        mlab.savefig(this_save_as)
                    else:
                        name_parts = this_save_as.split(".")
                        name_parts = name_parts[:-1] + [str(tind)] + name_parts[-1:]
                        frame_save_as = ".".join(name_parts)
                        mlab.savefig(frame_save_as)
                if plot_objects is None:
                    return plot_objects_to_return

            if animate:
                orig_offscreen_option = mlab.options.offscreen
                mlab.options.offscreen = True

                try:
                    # resets mlab.options.offscreen when it finishes, even if there is
                    # an error
                    if save_as is None:
                        raise ValueError(
                            "Must pass `save_as` for a mayavi animation, or no output will "
                            "be created"
                        )
                    with TemporaryDirectory() as d:
                        nframes = da.sizes[tcoord]

                        # First create png files in the temporary directory
                        temp_path = Path(d)
                        temp_save_as = str(temp_path.joinpath("temp.png"))
                        print("tind=0")
                        plot_objects = create_or_update_plot(
                            tind=0, this_save_as=temp_save_as
                        )

                        # @mlab.animate # interative mayavi animation too slow
                        def animation_func():
                            for tind in range(1, nframes):
                                print(f"tind={tind}")
                                create_or_update_plot(
                                    plot_objects=plot_objects,
                                    tind=tind,
                                    this_save_as=temp_save_as,
                                )
                                # yield # needed for an interactive mayavi animation

                        a = animation_func()

                        # Use ImageMagick via the wand package to turn the .png files into
                        # an animation
                        try:
                            from wand.image import Image
                        except ImportError:
                            raise ImportError(
                                "Please install the `wand` package to save the 3d animation"
                            )
                        with Image() as animation:
                            for i in range(nframes):
                                filename = str(temp_path.joinpath(f"temp.{i}.png"))
                                with Image(filename=filename) as frame:
                                    animation.sequence.append(frame)
                            animation.type = "optimize"
                            animation.loop = 0
                            # Delay is in milliseconds
                            animation.delay = int(round(1000.0 / fps))
                            animation.save(filename=save_as)
                finally:
                    mlab.options.offscreen = orig_offscreen_option

                # mlab.show() # interative mayavi animation so slow that it's not useful
                return a
            else:
                create_or_update_plot(this_save_as=save_as)

        else:
            raise ValueError(f"style='{style}' not implemented for engine='mayavi'")

        plt.show()
    else:
        raise ValueError(f"Unrecognised plot3d() 'engine' argument: {engine}")
